////////////////////////////////////////////////////////////////////////////////
// Filename  : interupt.c
// File Desc : Interrupt Handling
// Project   : NSFE (Nintendo Sixty Four Emulator)
// Release   : 0.2a (Developer Release)
// Copyright : NSFE is the Copyright (c) 1999, Brian Faber.  
//
// This software is a Windows implementation of TrueReality, Copyright (C) 1999,
// Niki W. Waibel, and is published under the terms of the GNU General Public 
// License, see below:
// -----------------------------------------------------------------------------
// This program is free software; you can redistribute it and or modify it under
// the terms of the GNU General Public Licence as published by the Free Software
// Foundation; either version 2 of the Licence, or any later version.
//
// This program is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU General Public Licence for more details
//
// You should have received a copy of the GNU General Public Licence along with 
// this program; if not, write to the Free Software Foundation, Inc., 675 Mass 
// Ave, Cambridge, MA 02139, USA.
//
// NSFE Author Information:
//    Brian Faber, Chino, Califorina - USA
//    bevis590@gte.net
//
//    Ben Marsh, Townsville, Queensland - Australia
//    photi@ozemail.com.au
//
// Original TrueReality Author Information:
//    Niki W. Waibel, Reichenau 20, 6890 Lustenau, Austria - EUROPE
//    niki.waibel@xms.net
// -----------------------------------------------------------------------------
// Update History
//
// 00001 16-02-1998 "Initial Code Release.                                     "

#ifndef _INTERUPT_H_
#include "interupt.h"
#endif // _INTERUPT_H_


////////////////////////////////////////////////////////////////////////////////
// Initialise Interrupts

   void InitInterrupts( void )
   {
      Interrupts[2].intrTime    = 0;
      Interrupts[2].intrHandler = MIInterruptVI;
  
      Interrupts[7].intrTime    = 0;
      Interrupts[7].intrHandler = CompareInterrupt;
   }


////////////////////////////////////////////////////////////////////////////////
// Check MI Interrupts

   void CheckMIInterrupts( void )
   {
      if( miOldIntrMaskReg != *(miIntrMaskReg) )
      {
         miOldIntrMaskReg = *(miIntrMaskReg);

         if( *(miIntrMaskReg) & *(miIntrReg) & MI_INTR_VI )
         {
            MIInterruptVI();
            return;
         }

         if( *(miIntrMaskReg) & *(miIntrReg) & MI_INTR_SI )
         {
            MIInterruptSI();
            return;
         }
      }
   }


////////////////////////////////////////////////////////////////////////////////
// Check Core CPU Interrupts

   void CheckCPUInterrupts( void )
   {
      // If any Interrupt without Mask Pending

      if( cp0Reg[C0_SR] & cp0Reg[C0_CAUSE] & SR_IMASK )   
      {
         // If the IE Bit in STATUS Register is Set
      
         if( cp0Reg[C0_SR] & SR_IE )
         {
            // If EXL or ERL Bits in STATUS Register are Clear
            
            if( (cp0Reg[C0_SR] & (SR_ERL | SR_EXL)) == SR_IMASK8 )
               CommonExceptionVector();
         }
      }
   }


////////////////////////////////////////////////////////////////////////////////
// Compare Interrupt (COUNT == COMPARE)

   void CompareInterrupt( void )
   {  
      // Set IP in CAUSE Register

      cp0Reg[C0_CAUSE] |= CAUSE_IP8;
   
      // If IP7 or IE Bits in STATUS Register are Clear then Return

      if( (cp0Reg[C0_SR] & (SR_IMASK7 | SR_IE)) != (SR_IMASK7 | SR_IE) )
         return;

      // If EXL or ERL Bits in STATUS Register are Set then Return

      if( cp0Reg[C0_SR] & (SR_EXL | SR_ERL) )
         return;

      CommonExceptionVector();
}


////////////////////////////////////////////////////////////////////////////////
// VI Interrupt

   void MIInterruptVI( void )
   {  
      Interrupts[2].intrTime += VI_INTR_TIME;

      // Add MI_INTR_VI in MI_INTR_REG

      *(miIntrReg) |= MI_INTR_VI;

      // If MI_INTR_VI in MI_INTR_MASK_REG is Disabled then Return

      if( !(*(miIntrMaskReg) & MI_INTR_VI) )
         return;

      // Set IP in CAUSE Register

      cp0Reg[C0_CAUSE] |= CAUSE_IP3;
   
      // If IP or IE Bits in STATUS Register are Clear then Return

      if( (cp0Reg[C0_SR] & (SR_IBIT3 | SR_IE)) != (SR_IBIT3 | SR_IE) )
         return;
   
      // If EXL or ERL Bits in STATUS Register are Set then Return

      if( cp0Reg[C0_SR] & (SR_ERL | SR_EXL) )
         return;

      CommonExceptionVector();
   } 


////////////////////////////////////////////////////////////////////////////////
// SI Interrupt
   
   void MIInterruptSI( void )
   {  
      // Add MI_INTR_SI in MI_INTR_REG

      *(miIntrReg) |= MI_INTR_SI;
   
      // If MI_INTR_SI in MI_INTR_MASK_REG is Disabled then Return

      if( !(*(miIntrMaskReg) & MI_INTR_SI) )
         return;

      // Set IP in CAUSE Register

      cp0Reg[C0_CAUSE] |= CAUSE_IP3;
   
      // If IP or IE Bits in STATUS Register are Clear then Return

      if( (cp0Reg[C0_SR] & (SR_IBIT3 | SR_IE)) != (SR_IBIT3 | SR_IE) )
         return;

      // If EXL or ERL Bits in STATUS Register are Set then Return

      if( cp0Reg[C0_SR] & (SR_ERL | SR_EXL) )
         return;

      CommonExceptionVector();
   }


////////////////////////////////////////////////////////////////////////////////
// Common Exception Vector Handler

   static void CommonExceptionVector( void )
   {
      // Set EXL in the STATUS Register

      cp0Reg[C0_SR] |= SR_EXL;   

      // Store PC in EPC

      cp0Reg[C0_EPC] = PC;
   
      if( branchDelay == 2 )
      {
         // Set BD Bit in CAUSE Register

         cp0Reg[C0_CAUSE] |= CAUSE_BD; 

         // Store PC before Branch Instruction

        cp0Reg[C0_EPC] -= 4;          
      } 

      // Set PC to 0x80000180 (Common Exception Vector)

      PC = E_VEC;

      // Next Instruction must NOT be the Instruction in the Delay Slot

      branchDelay = 0;
   }


