////////////////////////////////////////////////////////////////////////////////
// Filename  : memory.c
// File Desc : Emulated Memory
// Project   : NSFE (Nintendo Sixty Four Emulator)
// Release   : 0.2a (Developer Release)
// Copyright : NSFE is the Copyright (c) 1999, Brian Faber.  
//
// This software is a Windows implementation of TrueReality, Copyright (C) 1999,
// Niki W. Waibel, and is published under the terms of the GNU General Public 
// License, see below:
// -----------------------------------------------------------------------------
// This program is free software; you can redistribute it and or modify it under
// the terms of the GNU General Public Licence as published by the Free Software
// Foundation; either version 2 of the Licence, or any later version.
//
// This program is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU General Public Licence for more details
//
// You should have received a copy of the GNU General Public Licence along with 
// this program; if not, write to the Free Software Foundation, Inc., 675 Mass 
// Ave, Cambridge, MA 02139, USA.
//
// NSFE Author Information:
//    Brian Faber, Chino, Califorina - USA
//    bevis590@gte.net
//
//    Ben Marsh, Townsville, Queensland - Australia
//    photi@ozemail.com.au
//
// Original TrueReality Author Information:
//    Niki W. Waibel, Reichenau 20, 6890 Lustenau, Austria - EUROPE
//    niki.waibel@xms.net
// -----------------------------------------------------------------------------
// Update History
//
// 00001 16-02-1998 "Initial Code Release.                                     "

#ifndef _MEMORY_H_
#include "memory.h"
#endif // _MEMORY_H_//


////////////////////////////////////////////////////////////////////////////////
// Free Allocated Memory

   void FreeMem( void )
   {
      if( !GlobalSize( romImage ) )
         GlobalFree( romImage );

      if( !GlobalSize( rcpMem ) )
         GlobalFree( rcpMem );
   }


////////////////////////////////////////////////////////////////////////////////
// Allocated Memory

   BOOL InitMem( void )
   {
      // The Memory for the ROM Image is Allocated when the ROM is opened

      // Allocate RCP Memory/Register Memory

      rcpMem = (u32 *)GlobalAlloc( GMEM_FIXED | GMEM_ZEROINIT, MEMSIZE );

      if( rcpMem == NULL )
      {
         MessageBox( NULL, "Could not allocate memory for RCP memory.",
                              "Application Error", MB_OK );
	    	return( FALSE );
      }

      // Initialise ROM Pointers

      bootCode   = (romImage + 0x10);
      gameCode   = (romImage + 0x400);

      // Initialise RCP Memory/Register 'Block' Pointers

      rdramMem   = (rcpMem);
      rdramRegs  = (rcpMem + 0x104000);
      spMem      = (rcpMem + 0x12400A);
      spRegs     = (rcpMem + 0x12480A);
      dpCmdRegs  = (rcpMem + 0x13480C);
      dpSpanRegs = (rcpMem + 0x134814);
      miRegs     = (rcpMem + 0x134818);
      viRegs     = (rcpMem + 0x13481C);
      aiRegs     = (rcpMem + 0x13482A);
      piRegs     = (rcpMem + 0x134830);
      riRegs     = (rcpMem + 0x13483D);
      siRegs     = (rcpMem + 0x134845);
      pifRom     = (rcpMem + 0x13484C);
      pifRam     = (rcpMem + 0x134A3C);

      // Populate SP Memory (DMEM)

      CopyMemory( spMem, romImage, 0x1000 );

      // Initialise RCP Memory/Register Pointers

      // SP COP0 Registers

      spMemAddrReg      = (spRegs);
      spDramAddrReg     = (spRegs + 1);
      spRdLenReg        = (spRegs + 2);
      spWrLenReg        = (spRegs + 3);
      spStatusReg       = (spRegs + 4);
      spDmaFullReg      = (spRegs + 5);
      spDmaBusyReg      = (spRegs + 6);
      spSemaphoreReg    = (spRegs + 7);
	  //spPc              = (spRegs + 4);

      // DP Command Registers

      dpcStartReg       = (dpCmdRegs);          
      dpcEndReg         = (dpCmdRegs + 1);
      dpcCurrentReg     = (dpCmdRegs + 2);
      dpcStatusReg      = (dpCmdRegs + 3);
      dpcClockReg       = (dpCmdRegs + 4);
      dpcBufBusyReg     = (dpCmdRegs + 5);
      dpcPipeBusyReg    = (dpCmdRegs + 6);
      dpcTmemReg        = (dpCmdRegs + 7);

      // DP Span Registers

      dpsTbistReg       = (dpSpanRegs);
      dpsTestModeReg    = (dpSpanRegs + 1);
      dpsBufTestAddrReg = (dpSpanRegs + 2);
      dpsBufTestDataReg = (dpSpanRegs + 3);

      // MIPS Interface (MI) Registers 

      miInitModeReg     = (miRegs);
      miVersionReg      = (miRegs + 1);
      miIntrReg         = (miRegs + 2);
      miIntrMaskReg     = (miRegs + 3);

      // Video Interface (VI) Registers

      viStatusReg       = (viRegs);
      viOriginReg       = (viRegs + 1);
      viWidthReg        = (viRegs + 2);
      viIntrReg         = (viRegs + 3);
      viCurrentReg      = (viRegs + 4);
      viBurstReg        = (viRegs + 5);
      viVSyncReg        = (viRegs + 6);
      viHSyncReg        = (viRegs + 7);
      viLeapReg         = (viRegs + 8);
      viHStartReg       = (viRegs + 9);
      viVStartReg       = (viRegs + 10);
      viVBurstReg       = (viRegs + 11);
      viXScaleReg       = (viRegs + 12);
      viYScaleReg       = (viRegs + 13);

      // Audio Interface (AI) Registers

      aiDramAddrReg     = (aiRegs);
      aiLenReg          = (aiRegs + 1);
      aiControlReg      = (aiRegs + 2);
      aiStatusReg       = (aiRegs + 3);
      aiDacRateReg      = (aiRegs + 4);
      aiBitRateReg      = (aiRegs + 5);

      // Peripheral Interface (PI) Registers

      piDramAddrReg     = (piRegs);
      piCartAddrReg     = (piRegs + 1);
      piRdLenReg        = (piRegs + 2);
      piWrLenReg        = (piRegs + 3);
      piStatusReg       = (piRegs + 4);
      piBsdDom1LatReg   = (piRegs + 5);
      piBsdDom1PwdReg   = (piRegs + 6);
      piBsdDom1PgsReg   = (piRegs + 7);
      piBsdDom1RlsReg   = (piRegs + 8);
      piBsdDom2LatReg   = (piRegs + 9);
      piBsdDom2PwdReg   = (piRegs + 10);
      piBsdDom2PgsReg   = (piRegs + 11);
      piBsdDom2RlsReg   = (piRegs + 12);

      // RDRAM Interface (RI) Registers

      riModeReg         = (riRegs);            
      riConfigReg       = (riRegs + 1);
      riCurrentLoadReg  = (riRegs + 2);
      riSelectReg       = (riRegs + 3);
      riRefreshReg      = (riRegs + 4);
      riLatencyReg      = (riRegs + 5);
      riRerrorReg       = (riRegs + 6);
      riWerrorReg       = (riRegs + 7);

      // Serial Interface (SI) Registers

      siDramAddrReg     = (siRegs);
      siPifAddrRd64bReg = (siRegs + 1);
      siPifAddrWr64bReg = (siRegs + 4);
      siStatusReg       = (siRegs + 6);

      return( TRUE );
   }


////////////////////////////////////////////////////////////////////////////////
// Memory Access

   u32 *MemoryAccess( u32 vAddr, u32 flags )
   {
      u32 tAddr;

	  tAddr = (vAddr & 0x1FFFFFFF);

      switch( tAddr & 0x10000000 )
      {
         case 0x00000000:

            switch( tAddr & 0x0F000000 )
            {
               case 0x00000000:
               case 0x01000000:
               case 0x02000000:
               case 0x03000000:

                  // RDRAM Memory (plus RDRAM Control)
                  // 0x00000000 - 0x0040FFFF

                  if( tAddr < 0x40FFFF )
                     return( rdramMem + (tAddr / 4) );

                  // RDRAM Registers
                  // 0x03F00000 - 0x03F80027

                  if( tAddr >= 0x03F00000 )
                     return( rdramRegs + ((tAddr & 0x000FFFFF) / 4) );
                  
                  break;

               case 0x04000000:

                  switch( tAddr & 0x00F00000 )
                  {
                     case 0x00000000:
                        
                        switch( tAddr & 0x000F0000 )
                        {
                           case 0x00000000:

                              // SP Memory
                              // 0x04000000 - 0x04001FFF

                              if( tAddr < 0x04002000 )
                                 return( spMem + ((tAddr & 0xFFFF) / 4) );
							  break;

						   case 0x00040000:

                              // SP Registers
                              // 0x04040000 - 0x04080007

                              switch( tAddr )
                              {
                                 case SP_MEM_ADDR_REG:

                                    return( spMemAddrReg );

                                    break;

                                 case SP_DRAM_ADDR_REG:

                                    return( spDramAddrReg );

                                    break;

                                 case SP_RD_LEN_REG:

                                    return( spRdLenReg );

                                    break;

                                 case SP_WR_LEN_REG:

                                    return( spWrLenReg );

                                    break;

                                 case SP_STATUS_REG:

                                    return( spStatusReg );

                                    break;

                                 case SP_DMA_FULL_REG:

                                    return( spDmaFullReg );

                                    break;

                                 case SP_DMA_BUSY_REG:

                                    return( spDmaBusyReg );

                                    break;

                                 case SP_SEMAPHORE_REG:

                                    return( spSemaphoreReg );

                                    break;


                                 default:

                                    break;
                              }

							  case 0x00080000:

								  if(tAddr <= 0x04080004)
									 return( (u32 *)&nullMemory );
									 
								   break;
                        
                           default:
   
                              break;
                        }

                        break;
                        
                     case 0x00100000:

                        // DP Command Registers
                        // 0x04100000 - 0x0410001F

                        switch( tAddr )
                        {
                           case DPC_START_REG:

                              return( dpcStartReg );

                              break;

                           case DPC_END_REG:

                              return( dpcEndReg );

                              break;

                           case DPC_CURRENT_REG:

                              return( dpcCurrentReg );

                              break;

                           case DPC_STATUS_REG:

                              return( dpcStatusReg );

                              break;

                           case DPC_CLOCK_REG:

                              return( dpcClockReg );

                              break;

                           case DPC_BUFBUSY_REG:

                              return( dpcBufBusyReg );

                              break;

                           case DPC_PIPEBUSY_REG:

                              return( dpcPipeBusyReg );

                              break;

                           case DPC_TMEM_REG:

                              return( dpcTmemReg );

                              break;

                           default:

                              break;
                        }

                        break;
                        
                     case 0x00200000:

                        // DP Span Registers
                        // 0x04200000 - 0x0420000F

                        switch( tAddr )
                        {
                           case DPS_TBIST_REG:

                              return( dpsTbistReg );

                              break;

                           case DPS_TEST_MODE_REG:

                              return( dpsTestModeReg );

                              break;

                           case DPS_BUFTEST_ADDR_REG:

                              return( dpsBufTestAddrReg );

                              break;

                           case DPS_BUFTEST_DATA_REG:

                              return( dpsBufTestDataReg );

                              break;

                           default:
         
                              break;
                        }

                        break;
                            
                     case 0x00300000:

                        // MIPS Interface (MI) Registers
                        // 0x04300000 - 0x0430000F

                        switch( tAddr )
                        {
                           case MI_INIT_MODE_REG:

                              if( flags & MEMACCESS_WRITE )
                              {
                                 memAction |= UPDATE_MI_REG;
                                 cpuAction |= MI_INTERRUPTS;
                              }

                              return( miInitModeReg );

                              break;

                           case MI_VERSION_REG:

                              if( flags & MEMACCESS_WRITE )
                              {
                                 memAction |= UPDATE_MI_REG;
                                 cpuAction |= MI_INTERRUPTS;
                              }

                              return( miVersionReg );

                              break;

                           case MI_INTR_REG:

                              if( flags & MEMACCESS_WRITE )
                              {
                                 memAction |= UPDATE_MI_REG;
                                 cpuAction |= MI_INTERRUPTS;
                              }

                              return( miIntrReg );

                              break;

                           case MI_INTR_MASK_REG:

                              if( flags & MEMACCESS_WRITE )
                              {
                                 memAction |= UPDATE_MI_REG;
                                 cpuAction |= MI_INTERRUPTS;
                              }

                              return( miIntrMaskReg );

                              break;

                           default:

                              break;
                        }
                        
                        break;

                        
                     case 0x00400000:

                        // Video Interface (VI) Registers
                        // 0x04400000 - 0x04400037
   
                        switch( tAddr )
                        {
                           case VI_STATUS_REG:

                              if( flags & MEMACCESS_WRITE )
                                 memAction |= CHECK_VI_REG;

                              return( viStatusReg );

                              break;

                           case VI_ORIGIN_REG:

                              if( flags & MEMACCESS_WRITE )
                                 memAction |= CHECK_VI_REG;

                              return( viOriginReg );

                              break;

                           case VI_WIDTH_REG:

                              if( flags & MEMACCESS_WRITE )
                                 memAction |= CHECK_VI_REG;

                              return( viWidthReg );

                              break;

                           case VI_INTR_REG:

                              return( viIntrReg );

                              break;

                           case VI_CURRENT_REG:

                              if( flags & MEMACCESS_WRITE )
                              {
                                 // Clear VI Interrupt

                                   *(miIntrReg) &= ~MI_INTR_VI;
                              }

                              return( viCurrentReg );

                              break;

                           case VI_BURST_REG:

                              return( viBurstReg );

                              break;

                           case VI_V_SYNC_REG:

                              return( viVSyncReg );

                              break;

                           case VI_H_SYNC_REG:

                              return( viHSyncReg );

                              break;

                           case VI_LEAP_REG:

                              return( viLeapReg );

                              break;

                           case VI_H_START_REG:

                              return( viHStartReg );

                              break;

                           case VI_V_START_REG:

                              return( viVStartReg );

                              break;

                           case VI_V_BURST_REG:

                              return( viVBurstReg );

                              break;

                           case VI_X_SCALE_REG:

                              return( viXScaleReg );

                              break;

                           case VI_Y_SCALE_REG:

                              return( viYScaleReg );

                              break;

                           default:

                              break;
                        }

                        break;            
                        
                     case 0x00500000:

                        // Audio Interface (AI) Registers
                        // 0x04500000 - 0x04500017

                        switch( tAddr )
                        {
                           case AI_DRAM_ADDR_REG:

                              return( aiDramAddrReg );

                              break;

                           case AI_LEN_REG:

                              return( aiLenReg );

                              break;

                           case AI_CONTROL_REG:

                              return( aiControlReg );

                              break;

                           case AI_STATUS_REG:

                              return( aiStatusReg );

                              break;

                           case AI_DACRATE_REG:

                              return( aiDacRateReg );

                              break;

                           case AI_BITRATE_REG:

                              return( aiBitRateReg );

                              break;

                           default:

                              break;
                        }

                        break;
                            
                     case 0x00600000:

                        // Peripheral Interface (PI) Registers
                        // 0x04600000 - 0x04600033

                        switch( tAddr )
                        {
                           case PI_DRAM_ADDR_REG:

                              return( piDramAddrReg );

                              break;

                           case PI_CART_ADDR_REG:

                              return( piCartAddrReg );

                              break;

                           case PI_RD_LEN_REG:

                              return( piRdLenReg );

                              break;

                           case PI_WR_LEN_REG:

                              if( flags & MEMACCESS_WRITE )
                              {
                                 dmaFlags |= PI_DMA_WRITE;
                                 memAction |= PI_DMA_WRITE;
                              }

                              return( piWrLenReg );

                              break;

                           case PI_STATUS_REG:

							  if( flags & MEMACCESS_WRITE ) 
                                 memAction |= UPDATE_PI_REG;
							   
							  return( piStatusReg ); 

                              break;

                           case PI_BSD_DOM1_LAT_REG:

                              return( piBsdDom1LatReg );

                              break;

                           case PI_BSD_DOM1_PWD_REG:

                              return( piBsdDom1PwdReg );

                              break;

                           case PI_BSD_DOM1_PGS_REG:

                              return( piBsdDom1PgsReg );

                              break;

                           case PI_BSD_DOM1_RLS_REG:

                              return( piBsdDom1RlsReg );

                              break;

                           case PI_BSD_DOM2_LAT_REG:

                              return( piBsdDom2LatReg );

                              break;

                           case PI_BSD_DOM2_PWD_REG:

                              return( piBsdDom2PwdReg );

                              break;

                           case PI_BSD_DOM2_PGS_REG:

                              return( piBsdDom2PgsReg );

                              break;

                           case PI_BSD_DOM2_RLS_REG:

                              return( piBsdDom2RlsReg );

                              break;

                           default:

                              break;
                        }

                        break;
   
                     case 0x00700000:

                        // RDRAM Interface (RI) Registers
                        // 0x04700000 - 0x0470001F

                        switch( tAddr )
                        {
                           case RI_MODE_REG:

                              return( riModeReg );

                              break;

                           case RI_CONFIG_REG:

                              return( riConfigReg );

                              break;

                           case RI_CURRENT_LOAD_REG:

                              return( riCurrentLoadReg );

                              break;

                           case RI_SELECT_REG:

                              return( riSelectReg );

                              break;

                           case RI_REFRESH_REG:

                              return( riRefreshReg );

                              break;

                           case RI_LATENCY_REG:

                              return( riLatencyReg );

                              break;

                           case RI_RERROR_REG:

                              return( riRerrorReg );

                              break;

                           case RI_WERROR_REG:

                              return( riWerrorReg );

                              break;

                           default:

                              break;
                        }

                        break;
                              
                     case 0x00800000:

                        // Serial Interface (SI) Registers
                        // 0x04800000 - 0x0480001B

                        switch( tAddr )
                        {
                           case SI_DRAM_ADDR_REG:

                              return( siDramAddrReg );

                              break;

                           case SI_PIF_ADDR_RD64B_REG:

                              if( flags & MEMACCESS_WRITE )
                              {
                                 dmaFlags |= PIF2DRAM_DMA;
                                 memAction |= PIF2DRAM_DMA;
                              }

                              return( siPifAddrRd64bReg );

                              break;

                           case SI_PIF_ADDR_WR64B_REG:

                              if( flags & MEMACCESS_WRITE )
                              {
                                 dmaFlags |= DRAM2PIF_DMA;
                                 memAction |= DRAM2PIF_DMA;
                              }

                              return( siPifAddrWr64bReg );

                              break;

                           case SI_STATUS_REG:

                              if( flags & MEMACCESS_WRITE )
                              {
                                 // Clear SI Interrupt in MI Register

                                 *(miIntrReg) &= ~MI_INTR_SI; 
                                 *(siStatusReg) = 0; 
                              }

                              return( siStatusReg );

                              break;

                           default:

                              break;
                        }

                        break;

                     default:

                        break;
                  }

                  break;

               case 0x05000000:           
                      
                  // PI_DOM2_ADDR1

                  return( (u32 *)&nullMemory );

                  break;
                  
               case 0x06000000:

                  // PI_DOM1_ADDR1

                  if( tAddr < (0x06000000 + cartridgeSize) )
                     return( romImage + ((tAddr - 0x06000000) / 4) );

                  break;
                    
               case 0x08000000:

                  // PI_DOM2_ADDR2 

                  if( tAddr < (0x08000000 + cartridgeSize) )
                     return( romImage + ((tAddr - 0x08000000) / 4) );

                  break;
                  
               default:

                  break;

            } 

            break;
                
         case 0x10000000:

            // PI_DOM1_ADDR2

            if( tAddr < 0x1FC00000)                                  
            {
               if( tAddr < (0x10000000 + cartridgeSize) )
                  return( romImage + ((tAddr & 0xFFFFFFF) / 4) );

               break;
            }
                
            // PIF ROM

            if( tAddr < 0x1FC007C0 )                        
               return( pifRom + ((tAddr & 0xFFF) / 4) );

            // PIF RAM 

            if( tAddr < 0x1FC00800 )   
               return( pifRam + ((tAddr - 0x1FC007C0) / 4) );

            // PI_DOM1_ADDR3 

            if(tAddr < (0x1FD00000 + cartridgeSize))    
               return( romImage + ((tAddr - 0x1FD00000) / 4) );

            break;

		 /*case 0x80000000:
			 if(tAddr <= 0x9FFFFFFF)
			 {
				 //u32 tvAddr;
	             u32 trAddr;
				 switch( flags )
				 {
				 case MEMACCESS_READ:
				     trAddr = (s64)(*((s32 *)MemoryAccess(vAddr, MEMACCESS_READ)));
				     return((u32 *)&trAddr);
				     break;
				 case MEMACCESS_WRITE:
					 trAddr = (s64)(*((s32 *)MemoryAccess(vAddr, MEMACCESS_WRITE)));
					 return((u32 *)&trAddr);
					 break;
				 }
			 }
			 break;
		 case 0xA0000000:
			 if(tAddr <= 0xbFFFFFFF)
			 {
				 //u32 tvAddr;
	             u32 trAddr;
				 switch( flags )
				 {
				 case MEMACCESS_READ:
				     trAddr = (s64)(*((s32 *)MemoryAccess(vAddr, MEMACCESS_READ)));
				     return((u32 *)&trAddr);
				     break;
				 case MEMACCESS_WRITE:
					 trAddr = (s64)(*((s32 *)MemoryAccess(vAddr, MEMACCESS_WRITE)));
					 return((u32 *)&trAddr);
					 break;
				 }
			 }
			 break;*/


         
         default:

            break;
      }

      wsprintf( szBuffer, "Unmapped Memory Location\n\n"
                          "Virtual Address  <0x%08X>\n"
						  "Phisical Address <0x%08X>\n"
                          "Program Counter  <0x%08X>\n"
                          "CP0[Count] <0x%016X>\n\n"
                          "Action: Terminating Execution",
                          vAddr, tAddr, PC, cp0Reg[0x09] );

      if( flags & MEMACCESS_READ )
         MessageBox( hwndMain, szBuffer, "Emulated Memory Read Error", 
                     MB_OK | MB_ICONEXCLAMATION );
      else
         MessageBox( hwndMain, szBuffer, "Emulated Memory Write Error", 
                     MB_OK | MB_ICONEXCLAMATION );

      cpuIsRunning = FALSE;

      return( (u32 *)&nullMemory );
   }


////////////////////////////////////////////////////////////////////////////////
// Update MI Registers

   void UpdateMIReg( void )
   {
      // Set Mode Register for Reading
   
      *(miInitModeReg) &= (0xFFFFFF80 & *(miInitModeReg));
      *(miInitModeReg) |= (0x0000007F & *(miInitModeReg));
   
      if( *(miInitModeReg) & 0x0080 )        // Clear Init Mode
         *(miInitModeReg) &= 0xFFFFFF7F;
      if( *(miInitModeReg) & 0x0100 )        // Set Init Mode
         *(miInitModeReg) |= 0x00000080;
      if( *(miInitModeReg) & 0x0200 )        // Clear EBUS Test Mode
         *(miInitModeReg) &= 0xFFFFFEFF;
      if( *(miInitModeReg) & 0x0400 )        // Set EBUS Test Mode
         *(miInitModeReg) |= 0x00000100;
      if( *(miInitModeReg) & 0x1000 )        // Clear RDRAM Reg Mode
         *(miInitModeReg) &= 0xFFFFFDFF;
      if( *(miInitModeReg) & 0x2000 )        // Set RDRAM Reg Mode
         *(miInitModeReg) |= 0x00000200;
   
      // Set Version Register for Reading
   
      *(miVersionReg) = 0x01010101;
   
      // Set Interrupt Register for Reading
   
      if( *(miInitModeReg) & 0x0800 )
         *(miIntrReg) &= 0xFFFFFFDF;
   
      // Set Interrupt Mask Register for Reading
   
      if( *(miIntrMaskReg) & 0x0001 )        // Clear SP Interrupt Mask
         *(miIntrMaskReg) &= 0xFFFFFFFE;
      if( *(miIntrMaskReg) & 0x0002 )        // Set SP Interrupt Mask
         *(miIntrMaskReg) |= 0x00000001;
      if( *(miIntrMaskReg) & 0x0004 )        // Clear SP Interrupt Mask
         *(miIntrMaskReg) &= 0xFFFFFFFD;
      if( *(miIntrMaskReg) & 0x0008 )        // Set SP Interrupt Mask
         *(miIntrMaskReg) |= 0x00000002;
      if( *(miIntrMaskReg) & 0x0010 )        // Clear SP Interrupt Mask
         *(miIntrMaskReg) &= 0xFFFFFFFB;
      if( *(miIntrMaskReg) & 0x0020 )        // Set SP Interrupt Mask
         *(miIntrMaskReg) |= 0x00000004;
      if( *(miIntrMaskReg) & 0x0040 )        // Clear SP Interrupt Mask
         *(miIntrMaskReg) &= 0xFFFFFFF7;
      if( *(miIntrMaskReg) & 0x0080 )        // Set SP Interrupt Mask
         *(miIntrMaskReg) |= 0x00000008;
      if( *(miIntrMaskReg) & 0x0100 )        // Clear SP Interrupt Mask
         *(miIntrMaskReg) &= 0xFFFFFFEF;
      if( *(miIntrMaskReg) & 0x0200 )        // Set SP Interrupt Mask
         *(miIntrMaskReg) |= 0x00000010;
      if( *(miIntrMaskReg) & 0x0400 )        // Clear SP Interrupt Mask
         *(miIntrMaskReg) &= 0xFFFFFFDF;
      if( *(miIntrMaskReg) & 0x0800 )        // Set SP Interrupt Mask
         *(miIntrMaskReg) |= 0x00000020;

      // Set Registers for Writing
   
      *(miInitModeReg) = 0x00000000;
      *(miVersionReg)  = 0x01010101;
      *(miIntrReg)     = 0x00000000;
      *(miIntrMaskReg) = 0x00000000;
   }


////////////////////////////////////////////////////////////////////////////////
// Check VI Registers

   void CheckVIReg( void )
   {
      if( prevViStatusReg != *(viStatusReg) )
      {
         prevViStatusReg = *(viStatusReg);

         // Refresh Display

         return;
      }
        
      if( prevViOriginReg != *(viOriginReg) )
      {
         prevViOriginReg = *(viOriginReg);

         // Refresh Display

         return;
      }
        
      if( prevViWidthReg != *(viWidthReg) )
      {
         prevViWidthReg = *(viWidthReg);

         // Refresh Display

         return;
      }
   }


////////////////////////////////////////////////////////////////////////////////
// Update PI Registers

   void UpdatePIReg( void )
   {
      if( *(piStatusReg) & PI_STATUS_CLR_INTR )
         *(miIntrReg) &= ~MI_INTR_PI;
   }
