////////////////////////////////////////////////////////////////////////////////
// Filename  : n64.c
// File Desc : Nintendo 64 Core Emulation
// Project   : NSFE (Nintendo Sixty Four Emulator)
// Release   : 0.2a (Developer Release)
// Copyright : NSFE is the Copyright (c) 1999, Brian Faber.  
//
// This software is a Windows implementation of TrueReality, Copyright (C) 1999,
// Niki W. Waibel, and is published under the terms of the GNU General Public 
// License, see below:
// -----------------------------------------------------------------------------
// This program is free software; you can redistribute it and or modify it under
// the terms of the GNU General Public Licence as published by the Free Software
// Foundation; either version 2 of the Licence, or any later version.
//
// This program is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU General Public Licence for more details
//
// You should have received a copy of the GNU General Public Licence along with 
// this program; if not, write to the Free Software Foundation, Inc., 675 Mass 
// Ave, Cambridge, MA 02139, USA.
//
// NSFE Author Information:
//    Brian Faber, Chino, Califorina - USA
//    bevis590@gte.net
//
//    Ben Marsh, Townsville, Queensland - Australia
//    photi@ozemail.com.au
//
// Original TrueReality Author Information:
//    Niki W. Waibel, Reichenau 20, 6890 Lustenau, Austria - EUROPE
//    niki.waibel@xms.net
// -----------------------------------------------------------------------------
// Update History
//
// 00001 16-02-1998 "Initial Code Release.                                     "


#ifndef _N64_H_
#include "n64.h"
#endif // _N64_H_


////////////////////////////////////////////////////////////////////////////////
// Core Nintendo 64 Emulation Starts Here

   DWORD ExecNintendo64( LPVOID param )
   {
      s64 tValue64;                    // Temporary 64-Bit Value
      s32 tValue32;                    // Temporary 32-Bit Value
      u32 vAddr;                       // Virtual Address
      u64 tlbIndex;                    // TLB Entry Index
      
      // Main Emulation Loop

      while( cpuIsRunning )
      {
         // Cyclic Checks

         if( cpuAction )
         {
            // Check MI Interrupts

            if( cpuAction & MI_INTERRUPTS )
               CheckMIInterrupts();

            // Check the Core CPU Interrupts

            if( cpuAction & CPU_INTERRUPTS )
               CheckCPUInterrupts();

            // Clear Action Flags

            cpuAction = CLEAR_ACTIONS;
         }

         // Check whether a Compare Interrupt is Required

         if( cp0Reg[C0_COUNT] == cp0Reg[C0_COMPARE] )
            CompareInterrupt();

         // Check for a Timer Interrupt

         if( cp0Reg[C0_COUNT] == Interrupts[2].intrTime )
            ((void (*)())Interrupts[2].intrHandler)();

         // Get Instruction Word

         GetInstrWord();

         // Get Main Instruction Word Subdivisions
         // --------------------------------------
         // Bits [15:0]    Imm
         // Bits [20:16]   rt
         // Bits [25:21]   rs
         // Bits [31:26]   OpCode
                
         __asm {
            mov   eax,[instrWord]   // Get Immediate Value
            mov   [Imm],ax          
            mov   ecx,eax           // Get OpCode        
            shr   ecx,1Ah
            mov   [OpCode],cl
            mov   edx,eax           // Get 'rs' Value
            shr   edx,15h
            and   edx,1Fh
            mov   [rs],dl
            shr   eax,10h           // Get 'rt' Value
            and   eax,1Fh
            mov   [rt],al
         };

         // Debugger Functionality

         if( inDebugger )
         {
            SendMessage( hwndDebugGPRegs,   WM_COMMAND, ID_REFRESH, 0L );
            SendMessage( hwndCP0Regs,       WM_COMMAND, ID_REFRESH, 0L );
            SendMessage( hwndCP1Regs,       WM_COMMAND, ID_REFRESH, 0L );
            SendMessage( hwndCP2RegsCore,   WM_COMMAND, ID_REFRESH, 0L );
            SendMessage( hwndCP2RegsVector, WM_COMMAND, ID_REFRESH, 0L );
            SendMessage( hwndViAiDpsSiRegs, WM_COMMAND, ID_REFRESH, 0L );
            SendMessage( hwndPiRiRegs,      WM_COMMAND, ID_REFRESH, 0L );
            SendMessage( hwndDisassembly,   WM_COMMAND, ID_REFRESH, 0L );

            SuspendThread( hMainThread );
         }

         // Interpret OpCode

         switch( OpCode )
         {
            case SPECIAL:        // Special Instructions

               // Ignore NOP Instructions

               if( !instrWord )
                  break;

               // Get 'rd' and 'Funct' and 'sa' Subdivisions
               // ------------------------------------------
               // Bits [5:0]     Funct
               // Bits [10:6]    sa
               // Bits [15:11]   rd

               __asm {
                  mov   eax,[instrWord]   // Get 'rd' Value
                  mov   ecx,eax
                  shr   ecx,0Bh
                  and   ecx,1Fh
                  mov   [rd],cl
                  mov   edx,eax           // Get 'sa' Value
                  shr   edx,06h
                  and   edx,1Fh
                  mov   [sa],dl
                  and   eax,3Fh           // Get 'Funct' Value
                  mov   [Funct],al
               };

      			switch ( Funct )
               {
                  case SLL:  
                     
                     gpReg[rd] = (s64)(((u32)gpReg[rt]) << sa);

                     break;

                  case SRL:  

                     gpReg[rd] = (s64)(s32)(((u32)(gpReg[rt])) >> sa);

                     break;

                  case SRA:        

                     gpReg[rd] = (s64)((s32)gpReg[rt] >> sa);

                     break;

                  case SLLV:       

                     gpReg[rd] = (s64)((s32)((gpReg[rt]) << (gpReg[rs]&0x1F)));

                     break;

                  case SRLV:   

                     gpReg[rd] = (s64)(s32)((u32)gpReg[rt] >> (gpReg[rs]&0x1F));

                     break;

                  case SRAV: 

                     gpReg[rd] = (s64)((s32)gpReg[rt] >> (gpReg[rs] & 0x1F));

                     break;

                  case JR:   

                     branchPC = (u32)(gpReg[rs]);
                     branchDelay = 1;

                     break;

                  case JALR:         

					      branchDelay = 1;
					      branchPC = (u32)gpReg[rs];
					      gpReg[rd] = (s64)(PC + 8);

                     break;

                  case SYSCALL:      

                     ErrUnhandledOpCode( "SYSCALL" );

                     break;

                  case BREAK:          

                     //ErrUnhandledOpCode( "BREAK" );

					 // Ignor for now

                     break;

                  case SYNC:          

                     // Not Needed for Emulation.

                     break;

                  case MFHI:           

                     gpReg[rd] = HI;

                     break;

                  case MTHI:           

                     HI = gpReg[rs];

                     break;

                  case MFLO:           

                     gpReg[rd] = LO;

                     break;

                  case MTLO:           

                     LO = gpReg[rs];

                     break;

                  case DSLLV:         

                     gpReg[rd] = gpReg[rt] << (gpReg[rs] & 0x1F);
   
                     break;

                  case DSRLV:         

                     gpReg[rd] = (u64)gpReg[rt] >> (gpReg[rs] & 0x1F);

                     break;

                  case DSRAV:         

                     gpReg[rd] = gpReg[rt] >> (gpReg[rs] & 0x1F);

                     break;

                  case MULT:
                     
                     tValue64 = (s64)((s32)gpReg[rs]) * ((s32)gpReg[rt]);
                     LO = (s64)(s32)(tValue64);
                     HI = (s64)(s32)(tValue64 >> 32);

                     break;

                  case MULTU:

                     tValue64 = (u64)((u32)gpReg[rs]) * ((u32)gpReg[rt]);
                     LO = (s64)(s32)(tValue64);
                     HI = (s64)(s32)(tValue64 >> 32);

                     break;

                  case DIV:  

                     if( (s32)gpReg[rt] != 0 )
                     {
                        LO = (s64)(s32)((s32)gpReg[rs] / (s32)gpReg[rt]);
                        HI = (s64)(s32)((s32)gpReg[rs] % (s32)gpReg[rt]);
                     }

                     break;

                  case DIVU: 

                     if( (u32)gpReg[rt] != 0 )
                     {
                        LO = (s64)(s32)((u32)gpReg[rs] / (u32)gpReg[rt]);
                        HI = (s64)(s32)((u32)gpReg[rs] % (u32)gpReg[rt]);
                     }

                     break;

                  case DMULT:         

                     LO = gpReg[rs] * gpReg[rt];
                     HI = 0;

                     break;

                  case DMULTU:        

                     LO = (u64)(gpReg[rs]) * (u64)(gpReg[rt]);
                     HI = 0;

                     break;

                  case DDIV:          

                     if( gpReg[rt] != 0 )
                     {
                        LO = gpReg[rs] / gpReg[rt];
                        HI = gpReg[rs] % gpReg[rt];
                     }

                     break;

                  case DDIVU:      

                     if( (u64)gpReg[rt] != 0 )
                     {
                        LO = (u64)gpReg[rs] / (u64)gpReg[rt];
                        HI = (u64)gpReg[rs] % (u64)gpReg[rt];
                     }

                     break;

                  case ADD:           

                     gpReg[rd] = (s64)(s32)(gpReg[rs] + gpReg[rt]);

                     break;

                  case ADDU:          

                     gpReg[rd] = (s64)(s32)(gpReg[rs] + gpReg[rt]);

                     break;

                  case SUB:            

                     gpReg[rd] = (s64)(s32)(gpReg[rs] - gpReg[rt]);

                     break;

                  case SUBU:           

                     gpReg[rd] = (s64)((s32)gpReg[rs] - (s32)gpReg[rt]);

                     break;

                  case AND:            

                     gpReg[rd] = gpReg[rs] & gpReg[rt];

                     break;

                  case OR:             

                     gpReg[rd] = gpReg[rs] | gpReg[rt];

                     break;

                  case XOR:            

                     gpReg[rd] = gpReg[rs] ^ gpReg[rt];

                     break;

                  case NOR:            

                     gpReg[rd] = !(gpReg[rs] | gpReg[rt]);

                     break;

                  case SLT:            

                     if( gpReg[rs] < gpReg[rt] )
                        gpReg[rd] = 1;
                     else
                        gpReg[rd] = 0;

                     break;
                     
                  case SLTU:           

                     if( ((u64)(gpReg[rs])) < ((u64)(gpReg[rt])) )
                        gpReg[rd] = 1;
                     else
                        gpReg[rd] = 0;

                     break;

                  case DADD:         

                     gpReg[rd] = gpReg[rs] + gpReg[rt];

                     break;

                  case DADDU:        

                     gpReg[rd] = gpReg[rs] + gpReg[rt];

                     break;

                  case DSUB:         

                     gpReg[rd] = gpReg[rs] - gpReg[rt];

                     break;

                  case DSUBU:        

                     gpReg[rd] = gpReg[rs] - gpReg[rt];

                     break;

                  case TGE:          

                     ErrUnhandledOpCode( "TGE" );

                     break;

                  case TGEU:         

                     ErrUnhandledOpCode( "TGEU" );

                     break;

                  case TLT:          

                     ErrUnhandledOpCode( "TLT" );

                     break;

                  case TLTU:         

                     ErrUnhandledOpCode( "TLTU" );

                     break;

                  case TEQ:          

                     ErrUnhandledOpCode( "TEQ" );

                     break;

                  case TNE:          

                     ErrUnhandledOpCode( "TNE" );

                     break;

                  case DSLL:         

                     gpReg[rd] = gpReg[rt] << sa;
   
                     break;

                  case DSRL:         

                     gpReg[rd] = (u64)gpReg[rt] >> sa;
           
                     break;

                  case DSRA:         

                     gpReg[rd] = gpReg[rt] >> sa;
                     
                     break;

                  case DSLL32:       

                     gpReg[rd] = gpReg[rt] << (32 + sa);

                     break;

                  case DSRL32:       

                     gpReg[rd] = (u64)gpReg[rt] >> (32 + sa);
   
                     break;

                  case DSRA32:       

                     gpReg[rd] = gpReg[rt] >> (32 + sa);

                     break;

                  default:

                     // Unhandled Special OpCode

                     ErrUnhandledOpCode( "SPECIAL" );

                     break;
               };

               break;

            case REGIMM:         // Register Immediate Instructions

               switch( rt )
               {
                  case BLTZ:          

                     if( ((s64)gpReg[rs]) < 0 )
                     {
                        __asm {
                           mov   branchDelay,001h
                           movsx eax,Imm
                           shl   eax,002h
                           add   eax,PC
                           add   eax,04h
                           mov   branchPC,eax
                        };
                     }

                     break;

                  case BGEZ:           

                     if( ((s64)gpReg[rs]) >= 0 )
                     {
                        __asm {
                           mov   branchDelay,001h
                           movsx eax,Imm
                           shl   eax,002h
                           add   eax,PC
                           add   eax,04h
                           mov   branchPC,eax
                        };
                     }

                     break;

                  case BLTZL:         

                     if( ((s64)gpReg[rs]) < 0 )
                     {
                        __asm {
                           mov   branchDelay,001h
                           movsx eax,Imm
                           shl   eax,002h
                           add   eax,PC
                           add   eax,04h
                           mov   branchPC,eax
                        };
                     }
                     else
                        PC += 4;

                     break;

                  case BGEZL:         

                     if( ((s64)gpReg[rs]) >= 0 )
                     {
                        __asm {
                           mov   branchDelay,001h
                           movsx eax,Imm
                           shl   eax,002h
                           add   eax,PC
                           add   eax,04h
                           mov   branchPC,eax
                        };
                     }
                     else
                        PC += 4;

                     break;

                  case TGEIU:      

                     ErrUnhandledOpCode( "TGEIU" );

                     break;

                  case TLTI:       

                     ErrUnhandledOpCode( "TLTI" );

                     break;

                  case TLTIU:      

                     ErrUnhandledOpCode( "TLTIU" );

                     break;

                  case TEQI:       

                     ErrUnhandledOpCode( "TEQI" );

                     break;

                  case TNEI:       

                     ErrUnhandledOpCode( "TNEI" );

                     break;

                  case BLTZAL:        

                     gpReg[31] = PC + 8;

                     if( ((s64)gpReg[rs]) < 0 )
                     {
                        __asm {
                           mov   branchDelay,001h
                           movsx eax,Imm
                           shl   eax,002h
                           add   eax,PC
                           add   eax,04h
                           mov   branchPC,eax
                        };
                     }

                     break;

                  case BGEZAL:        

                     gpReg[0x1F] = PC + 8;

                     if( ((s64)gpReg[rs]) >= 0 )
                     {
                        __asm {
                           mov   branchDelay,001h
                           movsx eax,Imm
                           shl   eax,002h
                           add   eax,PC
                           add   eax,04h
                           mov   branchPC,eax
                        };
                     }

                     break;

                  case BLTZALL:     

                     gpReg[31] = PC + 8;

                     if( ((s64)gpReg[rs]) < 0 )
                     {
                        __asm {
                           mov   branchDelay,001h
                           movsx eax,Imm
                           shl   eax,002h
                           add   eax,PC
                           add   eax,04h
                           mov   branchPC,eax
                        };
                     }
                     else
                        PC += 4;

                     break;

                  case BGEZALL:   

                     gpReg[31] = PC + 8;

                     if( ((s64)gpReg[rs]) >= 0 )
                     {
                        __asm {
                           mov   branchDelay,001h
                           movsx eax,Imm
                           shl   eax,002h
                           add   eax,PC
                           add   eax,04h
                           mov   branchPC,eax
                        };
                     }
                     else
                        PC += 4;

                     break;

                  default:

                     // Unhandled Register Immediate OpCode

                     ErrUnhandledOpCode( "REGIMM" );

                     break;
               }

               break;

            case J:            

               __asm {
                  mov   branchDelay,001h
                  mov   eax,PC
                  and   eax,0F0000000h
                  mov   ebx,instrWord
                  and   ebx,003FFFFFFh
                  shl   ebx,002h
                  or    eax,ebx
                  mov   branchPC,eax
               };
 
               break;

            case JAL:          

               gpReg[31] = (s64)(s32)(PC + 8);
               __asm {
                  mov   branchDelay,001h
                  mov   eax,PC
                  and   eax,0F0000000h
                  mov   ebx,instrWord
                  and   ebx,003FFFFFFh
                  shl   ebx,002h
                  or    eax,ebx
                  mov   branchPC,eax
               };

               break;

            case BEQ:          

               if( gpReg[rs] == gpReg[rt] )
               {
                  __asm {
                     mov   branchDelay,001h
                     movsx eax,Imm
                     shl   eax,002h
                     add   eax,PC
                     add   eax,04h
                     mov   branchPC,eax
                  };
               }

               break;

            case BNE:  

               if( gpReg[rs] != gpReg[rt] )
               {
                  __asm {
                     mov   branchDelay,001h
                     movsx eax,Imm
                     shl   eax,002h
                     add   eax,PC
                     add   eax,04h
                     mov   branchPC,eax
                  };
               }

               break;

            case BLEZ:          

               if( ((s64)gpReg[rs]) <= 0 )
               {
                  __asm {
                     mov   branchDelay,001h
                     movsx eax,Imm
                     shl   eax,002h
                     add   eax,PC
                     add   eax,04h
                     mov   branchPC,eax
                  };
               }

               break;

            case BGTZ:          

               if( ((s64)gpReg[rs]) > 0 )
               {
                  __asm {
                     mov   branchDelay,001h
                     movsx eax,Imm
                     shl   eax,002h
                     add   eax,PC
                     add   eax,04h
                     mov   branchPC,eax
                  };
               }

               break;

            case ADDI:          

               gpReg[rt] = (s64)(s32)(gpReg[rs] + (s32)(Imm));

               break;

            case ADDIU:

               gpReg[rt] = (s64)(s32)(gpReg[rs] + (s32)(Imm));

               break;

            case SLTI:    

               if( gpReg[rs] < (s64)(s32)(Imm) )
                  gpReg[rt] = 1;
               else
                  gpReg[rt] = 0;

               break;

            case SLTIU:   

               if( (u64)gpReg[rs] < (u64)Imm )
                  gpReg[rt] = 1;
               else
                  gpReg[rt] = 0;

               break;

            case ANDI: 

               gpReg[rt] = gpReg[rs] & (u16)Imm;

               break;

            case ORI:  

               gpReg[rt] = gpReg[rs] | (u16)Imm;

               break;

            case XORI: 

               gpReg[rt] = gpReg[rs] ^ (u16)Imm;

               break;

            case LUI:

               // Load Upper Immediate
               // LUI rt,immediate
               
               __asm {
                  mov   ax,Imm                        // Load Immediate
                  shl   eax,010h                      // Shift Left 16-Bits
                  cdq                                 // Sign Extend Result
                  xor   ecx,ecx                       // Clear Register
                  mov   cl,[rt]                       // Load 'rt'
                  mov   dword ptr[ecx*8+gpReg],eax    // Store Result (Lo Word)
                  mov   dword ptr[ecx*8+gpReg+4],edx  // Store Result (Hi Word)
               };

               // Exceptions: None

               break;

            case COP0: 

               // Get 'rd' and 'Funct' and 'sa' Subdivisions
               // ------------------------------------------
               // Bits [5:0]     Funct
               // Bits [10:6]    sa
               // Bits [15:11]   rd

               __asm {
                  mov   eax,[instrWord]   // Get 'rd' Value
                  mov   ecx,eax
                  shr   ecx,0Bh
                  and   ecx,1Fh
                  mov   [rd],cl
                  mov   edx,eax           // Get 'sa' Value
                  shr   edx,06h
                  and   edx,1Fh
                  mov   [sa],dl
                  and   eax,3Fh           // Get 'Funct' Value
                  mov   [Funct],al
               };

               switch( rs )
               {
                  case MF:    

                     gpReg[rt] = (s64)(s32)cp0Reg[rd];

                     break;
               
                  case DMF:   

                     gpReg[rt] = cp0Reg[rd];

                     break;

                  case MT:

                     // If the Compare Register is written to then clear the
                     // External Interrupt Level 8 Pending Flag

                     if( gpReg[rd] == C0_COMPARE )
                        cp0Reg[C0_CAUSE] &= ~CAUSE_IP8;
   
                     // If the Cause Register is written to then Mask out all
                     // the Bits except Interrupt Pending 0 and 1

                     if( gpReg[rd] == C0_CAUSE )
                     {
                        cp0Reg[C0_CAUSE] = gpReg[rt] & CAUSE_IP01MASK;
                        break;
                     }
   
                     cp0Reg[rd] = gpReg[rt];
   
                     // Check for Interrupts if Status Register Written to

                     if( (gpReg[rd] == C0_SR) && (gpReg[rt] & CAUSE_IPMASK) )
                        cpuAction |= CPU_INTERRUPTS;

                     break;

                  default:

                     switch( Funct )
                     {
                        case TLBR:

                           tlbIndex = cp0Reg[C0_INX];

                           if( tlbIndex <= NTLBENTRIES )
                           {
                              cp0Reg[C0_PAGEMASK] = tlb[tlbIndex].PageMask;
                              cp0Reg[C0_ENTRYHI]  = tlb[tlbIndex].EntryHi & 
                                                    (~tlb[tlbIndex].PageMask);
                              cp0Reg[C0_ENTRYLO1] = tlb[tlbIndex].EntryLo1;
                              cp0Reg[C0_ENTRYLO0] = tlb[tlbIndex].EntryLo0;
      
                              if( tlb[tlbIndex].G == 0 )
                              {  
                                 // Clear G Bit if not set in the TLB Entry
            
                                 cp0Reg[C0_ENTRYLO1] &= 0xFFFFFFFFFFFFFFFE;
                                 cp0Reg[C0_ENTRYLO0] &= 0xFFFFFFFFFFFFFFFE;
                              }
                              else
                              {
                                 // Otherwise Set G Bit
   
                                 cp0Reg[C0_ENTRYLO1] |= TLBLO_G;
                                 cp0Reg[C0_ENTRYLO0] |= TLBLO_G;
                              }
         
                           }

                           break;

                        case TLBWI:

                           tlbIndex = cp0Reg[C0_INX];

                           if( tlbIndex <= NTLBENTRIES )
                           {
                              tlb[tlbIndex].PageMask = cp0Reg[C0_PAGEMASK];
                              tlb[tlbIndex].EntryHi  = cp0Reg[C0_ENTRYHI] & 
                                                       (~cp0Reg[C0_PAGEMASK]);
                              tlb[tlbIndex].EntryLo1 = cp0Reg[C0_ENTRYLO1];
                              tlb[tlbIndex].EntryLo0 = cp0Reg[C0_ENTRYLO0];
                              tlb[tlbIndex].G        = TLBLO_G & 
                                                       cp0Reg[C0_ENTRYLO1] & 
                                                       cp0Reg[C0_ENTRYLO0];
                           }

                           break;

                        case TLBWR:

                           tlbIndex = cp0Reg[C0_RAND];

                           if( tlbIndex <= NTLBENTRIES )
                           {
                              tlb[tlbIndex].PageMask = cp0Reg[C0_PAGEMASK];
                              tlb[tlbIndex].EntryHi  = cp0Reg[C0_ENTRYHI] & 
                                                       (~cp0Reg[C0_PAGEMASK]);
                              tlb[tlbIndex].EntryLo1 = cp0Reg[C0_ENTRYLO1];
                              tlb[tlbIndex].EntryLo0 = cp0Reg[C0_ENTRYLO0];
                              tlb[tlbIndex].G        = TLBLO_G & 
                                                       cp0Reg[C0_ENTRYLO1] &
                                                       cp0Reg[C0_ENTRYLO0];
                           }

                           break;

                        case TLBP:

                           cp0Reg[C0_INX] = 0x80000000;

                           // 64-Bit Mode

                           if( cp0Reg[C0_SR] & 0xE0 )  
                           {
                              for( tlbIndex = 0; tlbIndex < 32; tlbIndex++ )
                              {
                                 if(
                                    (((tlb[tlbIndex].EntryHi >> 13)&0x7FFFFFF) &
                                    (~(tlb[tlbIndex].PageMask >> 13)))
                                    ==
                                    (((cp0Reg[C0_ENTRYHI] >> 13) & 0x7FFFFFF) &
                                    (~(tlb[tlbIndex].PageMask >> 13)))
                                    )
                                 {
                                    if( !tlb[tlbIndex].G )
                                    {
                                       if((tlb[tlbIndex].EntryHi & 0xFF) == 
                                          (cp0Reg[C0_ENTRYHI] & 0xFF))
                                       {  
                                          cp0Reg[C0_INX] = tlbIndex;
                                          break;
                                       }
                                    }
                                    else
                                    {
                                       cp0Reg[C0_INX] = tlbIndex;
                                       break;
                                    }
                                 }
                              }
                           }
                           else   // 32-Bit Mode
                           {
                              for( tlbIndex = 0; tlbIndex < 32; tlbIndex++ )
                              {
                                 if(
                                    (((tlb[tlbIndex].EntryHi >> 13) & 0x7FFFF) &
                                    (~(tlb[tlbIndex].PageMask >> 13)))
                                    ==
                                    (((cp0Reg[C0_ENTRYHI] >> 13) & 0x7FFFF) &
                                    (~(tlb[tlbIndex].PageMask >> 13)))
                                    )
                                 {
                                    if( !tlb[tlbIndex].G )
                                    {
                                       if((tlb[tlbIndex].EntryHi & 0xFF) ==
                                          (cp0Reg[C0_ENTRYHI] & 0xFF))
                                       {
                                          cp0Reg[C0_INX] = tlbIndex;
                                          break;
                                       }
                                    }
                                    else
                                    {
                                       cp0Reg[C0_INX] = tlbIndex;
                                       break;
                                    }
                                 }
                              }
                           }

                           break;

                        case ERET:

                           // If ERL Exception

                           if( cp0Reg[C0_SR] & SR_ERL )
                           {
                              PC = (u32)cp0Reg[C0_ERROR_EPC] - 4;
                              cp0Reg[C0_SR] &= 0xFFFFFFFFFFFFFFFB;
                           }
                           else     // Else "Normal" Exception
                           {
                              PC = (u32)cp0Reg[C0_EPC] - 4;
                              cp0Reg[C0_SR] &= 0xFFFFFFFFFFFFFFFD;
                           }

                           branchDelay = 0;
                           LLBit = 0;

                           break;

                        default:

                           // Unhandled COP0 OpCode

                           ErrUnhandledOpCode( "COP0" );

                           break;
                     };

                     break;
               };

               break;

            case COP1:

               // Get 'rd' and 'Funct' and 'sa' Subdivisions
               // ------------------------------------------
               // Bits [5:0]     Funct
               // Bits [10:6]    sa
               // Bits [15:11]   rd

               __asm {
                  mov   eax,[instrWord]   // Get 'rd' Value
                  mov   ecx,eax
                  shr   ecx,0Bh
                  and   ecx,1Fh
                  mov   [rd],cl
                  mov   edx,eax           // Get 'sa' Value
                  shr   edx,06h
                  and   edx,1Fh
                  mov   [sa],dl
                  and   eax,3Fh           // Get 'Funct' Value
                  mov   [Funct],al
               };

               switch( rs )
               {
                  case MF: 

                     gpReg[rt] = (s64)(s32)cp1Reg[rd];

                     break;
               
                  case CF: 

                     gpReg[rt] = (s64)cp1Reg[rd];

                     break;

                  case CT: 

                     cp1Reg[rd] = gpReg[rt];

                     break;

                  case MT: 
 
                     cp1Reg[rd] = gpReg[rt];

                     break;

                  default:

                     switch( Funct )
                     {
                        case ADDfmt:

                           switch( rs )
                           {
                              case FMT_S:

                                 *((float *)&cp1Reg[sa]) = 
                                    *((float *)&cp1Reg[rd]) + 
                                    *((float *)&cp1Reg[rt]);

                                 break;

                              case FMT_D:

                                 *((double *)&cp1Reg[sa]) = 
                                    *((double *)&cp1Reg[rd]) + 
                                    *((double *)&cp1Reg[rt]);

                                 break;

                              default:

                                 ErrUnhandledOpCode( "ADDfmt" );

                                 break;
                           }

                           break;

                        case MULfmt:

                           switch( rs )
                           {
                              case FMT_S:

                                 *((float *)&cp1Reg[sa]) = 
                                    *((float *)&cp1Reg[rd]) * 
                                    *((float *)&cp1Reg[rt]);

                                 break;

                              case FMT_D:

                                 *((double *)&cp1Reg[sa]) =
                                    *((double *)&cp1Reg[rd]) *
                                    *((double *)&cp1Reg[rt]);

                                 break;

                              default:

                                 ErrUnhandledOpCode( "MULfmt" );

                                 break;
                           }

                           break;

                        case DIVfmt:

                           switch( rs )
                           {
                              case FMT_S:

                                 if( *((float *)&cp1Reg[rt]) != 0 )
                                    *((float *)&cp1Reg[sa]) = 
                                       *((float *)&cp1Reg[rd]) / 
                                       *((float *)&cp1Reg[rt]);
                                 else
                                    // Division by Zero Exception
                                    ;

                                 break;

                              case FMT_D:

                                 if( *((float *)&cp1Reg[rt]) != 0 )
                                    *((double *)&cp1Reg[sa]) =
                                       *((double *)&cp1Reg[rd]) /
                                       *((double *)&cp1Reg[rt]);
                                 else
                                    // Division by Zero Exception
                                    ;

                                 break;

                              default:

                                 ErrUnhandledOpCode( "DIVfmt" );

                                 break;
                           }
 
                           break;

                        case TRUNCWfmt:

                           switch( rs )
                           {
                              case FMT_S:

                                 *((s32 *)&cp1Reg[sa]) = 
                                    (s32)*((float *)&cp1Reg[rd]);

                                 break;

                              case FMT_D:

                                 *((s32 *)&cp1Reg[sa]) =
                                    (s32)*((double *)&cp1Reg[rd]);

                                 break;

                              default:

                                 ErrUnhandledOpCode( "TRUNCWfmt" );

                                 break;
                           }
 
                           break;

                        case CVTSfmt:

                           switch( rs )
                           {
                              case FMT_D:

                                 *((float *)&cp1Reg[sa]) =
                                    (float)*((double *)&cp1Reg[rd]);

                                 break;

                              case FMT_W:

                                 *((float *)&cp1Reg[sa]) = 
                                    (float)(s32)cp1Reg[rd];

                                 break;

                              case FMT_L:

                                 *((float *)&cp1Reg[sa]) =
                                    (float)(s64)cp1Reg[rd];

                                 break;

                              default:

                                 ErrUnhandledOpCode( "CVTSfmt" );

                                 break;
                           }
 
                           break;

                        case CVTDfmt:

                           switch( rs )
                           {
                              case FMT_S:

                                 *((double *)&cp1Reg[sa]) =
                                    (double)*((float *)&cp1Reg[rd]);

                                 break;

                              case FMT_W:

                                 *((double *)&cp1Reg[sa]) = 
                                    (double)(s32)cp1Reg[rd];

                                 break;

                              case FMT_L:

                                 *((double *)&cp1Reg[sa]) = 
                                    (double)(s64)cp1Reg[rd];

                                 break;

                              default:

                                 ErrUnhandledOpCode( "CVTDfmt" );

                                 break;
                           }
 
                           break;

                        default:

                           // Unhandled COP1 OpCode

                           ErrUnhandledOpCode( "COP1" );

                           break;
                     }
                  }

               break;

            case COP2:

               // Get 'rd' and 'Funct' and 'sa' Subdivisions
               // ------------------------------------------
               // Bits [5:0]     Funct
               // Bits [10:6]    sa
               // Bits [15:11]   rd

               __asm {
                  mov   eax,[instrWord]   // Get 'rd' Value
                  mov   ecx,eax
                  shr   ecx,0Bh
                  and   ecx,1Fh
                  mov   [rd],cl
                  mov   edx,eax           // Get 'sa' Value
                  shr   edx,06h
                  and   edx,1Fh
                  mov   [sa],dl
                  and   eax,3Fh           // Get 'Funct' Value
                  mov   [Funct],al
               };

               switch( rs )
               {
                  case VECTOR_F:

                     switch( Funct )
                     {
                        case VMULF:

                           // Vector (Fraction) Multiply

                           ErrUnhandledOpCode( "VMULF" );

                           break;

                        case VMULU:

                           // Vector (Unsigned Fraction) Multiply

                           ErrUnhandledOpCode( "VMULU" );

                           break;

                        case VRNDP:

                           // Vector DCT Round (+)

                           ErrUnhandledOpCode( "VRNDP" );

                           break;

                        case VMULQ:

                           // Vector (Integer) Multiply

                           ErrUnhandledOpCode( "VMULQ" );

                           break;

                        case VMUDL:

                           // Vector (Low) Multiply

                           ErrUnhandledOpCode( "VMUDL" );

                           break;

                        case VMUDM:

                           // Vector (Mid-M) Multiply

                           ErrUnhandledOpCode( "VMUDM" );

                           break;

                        case VMUDN:

                           // Vector (Mid-N) Multiply

                           ErrUnhandledOpCode( "VMUDN" );

                           break;

                        case VMUDH:

                           // Vector (High) Multiply

                           ErrUnhandledOpCode( "VMUDH" );

                           break;

                        case VMACF:

                           // Vector (Fraction) Multiply Accumulate

                           ErrUnhandledOpCode( "VMACF" );

                           break;

                        case VMACU:

                           // Vector (Unsigned Fraction) Multiply Accumulate

                           ErrUnhandledOpCode( "VMACU" );

                           break;

                        case VRNDN:

                           // Vector DCT Round (-)

                           ErrUnhandledOpCode( "VRNDN" );

                           break;

                        case VMACQ:

                           // Vector (Integer) Multiply Accumulate

                           ErrUnhandledOpCode( "VMACQ" );

                           break;

                        case VMADL:

                           // Vector (Low) Multiply Accumulate

                           ErrUnhandledOpCode( "VMADL" );

                           break;

                        case VMADM:

                           // Vector (Mid-M) Multiply Accumulate

                           ErrUnhandledOpCode( "VMADM" );

                           break;

                        case VMADN:

                           // Vector (Mid-N) Multiply Accumulate

                           ErrUnhandledOpCode( "VMADN" );

                           break;

                        case VMADH:

                           // Vector (High) Multiply Accumulate

                           ErrUnhandledOpCode( "VMADH" );

                           break;

                        case VADD:

                           // Vector Add

                           ErrUnhandledOpCode( "VADD" );

                           break;

                        case VSUB:

                           // Vector Subtract

                           ErrUnhandledOpCode( "VSUB" );

                           break;

                        case VSUT:

                           // Vector SUT (vt - vs)

                           ErrUnhandledOpCode( "VSUT" );

                           break;

                        case VABS:

                           // Vector Absolute Value

                           ErrUnhandledOpCode( "VABS" );

                           break;

                        case VADDC:

                           // Vector ADDC

                           ErrUnhandledOpCode( "VADDC" );

                           break;

                        case VSUBC:

                           // Vector SUBC

                           ErrUnhandledOpCode( "VSUBC" );

                           break;

                        case VADDB:

                           // Vector Add Byte

                           ErrUnhandledOpCode( "VADDB" );

                           break;

                        case VSUBB:

                           // Vector Subtract Byte

                           ErrUnhandledOpCode( "VSUBB" );

                           break;

                        case VACCB:

                           // Vector Add Byte/Add Accumulator

                           ErrUnhandledOpCode( "VACCB" );

                           break;

                        case VSUCB:

                           // Vector Subtract Byte/Add Accumulator

                           ErrUnhandledOpCode( "VSUCB" );

                           break;

                        case VSAD:

                           // Vector SAD

                           ErrUnhandledOpCode( "VSAD" );

                           break;

                        case VSAC:

                           // Vector SAC

                           ErrUnhandledOpCode( "VSAC" );

                           break;

                        case VSUM:

                           // Vector SUM

                           ErrUnhandledOpCode( "VSUM" );

                           break;

                        case VSAW:

                           // Vector SAW

                           ErrUnhandledOpCode( "VSAW" );

                           break;

                        case VLT:

                           // Vector Less Than

                           ErrUnhandledOpCode( "VLT" );

                           break;

                        case VEQ:

                           // Vector Equal To

                           ErrUnhandledOpCode( "VEQ" );

                           break;

                        case VNE:

                           // Vector Not Equal To

                           ErrUnhandledOpCode( "VNE" );

                           break;

                        case VGE:

                           // Vector Greater Than or Equal To

                           ErrUnhandledOpCode( "VGE" );

                           break;

                        case VCL:

                           // Vector Clip Low

                           ErrUnhandledOpCode( "VCL" );

                           break;

                        case VCH:

                           // Vector Clip High

                           ErrUnhandledOpCode( "VCH" );

                           break;

                        case VCR:

                           // Vector, 1's Complement Clip

                           ErrUnhandledOpCode( "VCR" );

                           break;

                        case VMRG:

                           // Vector Merge

                           ErrUnhandledOpCode( "VMRG" );

                           break;

                        case VAND:

                           // Vector Logical AND

                           ErrUnhandledOpCode( "VAND" );

                           break;

                        case VNAND:

                           // Vector Logical NAND

                           ErrUnhandledOpCode( "VNAND" );

                           break;

                        case VOR:

                           // Vector Logical OR

                           ErrUnhandledOpCode( "VOR" );

                           break;

                        case VNOR:

                           // Vector Logical NOR

                           ErrUnhandledOpCode( "VNOR" );

                           break;

                        case VNXOR:

                           // Vector Logical NXOR

                           ErrUnhandledOpCode( "VNXOR" );

                           break;

                        case VEXTT:

                           // Vector Extract Triple (5/5/5/1)

                           ErrUnhandledOpCode( "VEXTT" );

                           break;

                        case VEXTQ:

                           // Vector Extract Quad (4/4/4/4)

                           ErrUnhandledOpCode( "VEXTQ" );

                           break;

                        case VEXTN:

                           // Vector Extract Nibble (4/4/4/4) Sign-Extended

                           ErrUnhandledOpCode( "VEXTN" );

                           break;

                        case VINST:

                           // Vector Insert Triple (5/5/5/1)

                           ErrUnhandledOpCode( "VINST" );

                           break;

                        case VINSQ:

                           // Vector Insert Quad (4/4/4/4)

                           ErrUnhandledOpCode( "VINSQ" );

                           break;

                        case VINSN:

                           // Vector Insert Nibble (4/4/4/4) Sign-Extended

                           ErrUnhandledOpCode( "VINSN" );

                           break;

                        default:

                           // Unhandled 'Funct' Vector Instruction

                           ErrUnhandledOpCode( "Vector Function" );

                           break;
                     }

                     break;

                  default:

                     // Unhandled COP2 OpCode

                     ErrUnhandledOpCode( "COP2" );

                     break;
               }

               break;

            case BEQL:                

               if( gpReg[rs] == gpReg[rt] )
               {
                  __asm {
                     mov   branchDelay,001h
                     movsx eax,Imm
                     shl   eax,002h
                     add   eax,PC
                     add   eax,04h
                     mov   branchPC,eax
                  };
               }
               else
                  PC += 4;

               break;

            case BNEL:    

               if( gpReg[rs] != gpReg[rt] )
               {
                  __asm {
                     mov   branchDelay,001h
                     movsx eax,Imm
                     shl   eax,002h
                     add   eax,PC
                     add   eax,04h
                     mov   branchPC,eax
                  };
               }
               else
                  PC += 4;

               break;

            case BLEZL:           

               if( ((s64)(gpReg[rs])) <= 0 )
               {
                  __asm {
                     mov   branchDelay,001h
                     movsx eax,Imm
                     shl   eax,002h
                     add   eax,PC
                     add   eax,04h
                     mov   branchPC,eax
                  };
               }
               else
                  PC += 4;

               break;

            case BGTZL:           
  
               if( ((s64)gpReg[rs]) > 0 )
               {
                  __asm {
                     mov   branchDelay,001h
                     movsx eax,Imm
                     shl   eax,002h
                     add   eax,PC
                     add   eax,04h
                     mov   branchPC,eax
                  };
               }
               else
                  PC += 4;

               break;

            case DADDI:           

               gpReg[rt] = gpReg[rs] + Imm;

               break;

            case DADDIU:          

               gpReg[rt] = gpReg[rs] + Imm;

               break;

            case LDL:             

               ErrUnhandledOpCode( "LDL" );

               break;

            case LDR:             

               ErrUnhandledOpCode( "LDR" );

               break;

            case LB:   

               __asm {
                  movsx eax,Imm
                  mov   tValue32,eax
               };

               vAddr = (u32)(((s64)(tValue32)) + gpReg[rs]);
               vAddr ^= 0x03;

               __asm {
                  mov   eax,vAddr
                  and   eax,000000003h
                  cmp   eax,00h
                  jne   lb_addrexcept
               };

               gpReg[rt] = (s64)(*((s8 *)MemoryAccess(vAddr, MEMACCESS_READ)));

               break;

               lb_addrexcept:

                  wsprintf(szBuffer,"Address Error Exception at <0x%08X>.",PC);
                  MessageBox( hwndMain, szBuffer, 
                              "Load Byte", MB_OK | MB_ICONEXCLAMATION );
                  break;

            case LH:      

               ErrUnhandledOpCode( "LH" );

               break;

            case LWL:     

               ErrUnhandledOpCode( "LWL" );

               break;

            case LW: 

               __asm {
                  movsx eax,Imm
                  mov   tValue32,eax
               };

               vAddr = (u32)(((s64)(tValue32)) + gpReg[rs]);

               __asm {
                  mov   eax,vAddr
                  and   eax,000000003h
                  cmp   eax,00h
                  jne   lw_addrexcept
               };

               gpReg[rt] = (s64)(*((s32 *)MemoryAccess(vAddr, MEMACCESS_READ)));

               break;

               lw_addrexcept:

                  wsprintf(szBuffer,"Address Error Exception at <0x%08X>.",PC);
                  MessageBox( hwndMain, szBuffer, 
                              "Load Word", MB_OK | MB_ICONEXCLAMATION );
                  break;

            case LBU:             

               __asm {
                  movsx eax,Imm
                  mov   tValue32,eax
               };

               vAddr = (u32)(((s64)(tValue32)) + gpReg[rs]);
               vAddr ^= 0x03;

               __asm {
                  mov   eax,vAddr
                  and   eax,000000003h
                  cmp   eax,00h
                  jne   lbu_addrexcept
               };

               gpReg[rt] = (u64)(*((u8 *)MemoryAccess(vAddr, MEMACCESS_READ)));

               break;

               lbu_addrexcept:

                  wsprintf(szBuffer,"Address Error Exception at <0x%08X>.",PC);
                  MessageBox( hwndMain, szBuffer, 
                              "Load Byte Unsigned", MB_OK | MB_ICONEXCLAMATION);
                  break;

            case LHU:             

               //ErrUnhandledOpCode( "LHU" );

               __asm {
                  movsx eax,Imm
                  mov   tValue32,eax
               };

               vAddr = (UINT32)(((INT64)(tValue32)) + gpReg[rs]);

               // TO DO: Address Error Exception
               //        TLB Refill Exception
               //        TLB Invalid Exception
               //        Bus Error Exception

               //if( (vAddr & 0x03) )
               //{
               //   tValue32 = *( 
               //      PhysicalAddress( hwnd, vAddr ^ 0x02, MEM_READ, MEM_HWORD ) );

               //      gpReg[rt] = (UINT64)(tValue32 & 0x0000FFFF);
               //}
               //else
               //{
                  tValue32 = *(MemoryAccess(vAddr, MEMACCESS_READ));

                  gpReg[rt] = (UINT64)((tValue32 & 0xFFFF0000) >> 16);
               //}

               break;

            case LWR:             

               ErrUnhandledOpCode( "LWR" );

               break;

            case LWU:             

               ErrUnhandledOpCode( "LWU" );

               break;

            case SB:              

               __asm {
                  movsx eax,Imm
                  mov   tValue32,eax
               };

               vAddr = (u32)(((s64)(tValue32)) + gpReg[rs]);
               vAddr ^= 0x03;

               __asm {
                  mov   eax,vAddr
                  and   eax,000000003h
                  cmp   eax,00h
                  jne   sb_addrexcept
               };

               *((s32 *)MemoryAccess( vAddr, MEMACCESS_WRITE )) &= 0xFFFFFF00;
               *((s32 *)MemoryAccess( vAddr, MEMACCESS_WRITE )) += 
                  ((s32)(gpReg[rt] & 0xFF));

               MemWriteCheck();

               break;

               sb_addrexcept:

                  wsprintf(szBuffer,"Address Error Exception at <0x%08X>.",PC);
                  MessageBox( hwndMain, szBuffer, 
                              "Store Byte", MB_OK | MB_ICONEXCLAMATION );
                  break;

            case SH:           

               //ErrUnhandledOpCode( "SH" );
               // MemWriteCheck();

               __asm {
                  movsx eax,Imm
                  mov   tValue32,eax
               };

               vAddr = (UINT32)(((INT64)(tValue32)) + gpReg[rs]);

               __asm {
                  mov   eax,vAddr
                  and   eax,000000003h
                  cmp   eax,00h
                  jne   sh_addrexcept
               };

               tValue32 =  *((s32 *)MemoryAccess(vAddr, MEMACCESS_READ));
               tValue32 &= 0x0000FFFF;
               tValue32 |= (INT32)((gpReg[rt] & 0xFFFF)  << 16);

               *((s32 *)MemoryAccess(vAddr, MEMACCESS_WRITE)) = tValue32;

               MemWriteCheck();

               break;

               sh_addrexcept:

                  // TO DO: Address Error Exception
                  //        TLB Refill Exception
                  //        TLB Modification Exception
                  //        TLB Invalid Exception
                  //        Bus Error Exception

                  break;

            case SWL:          

               ErrUnhandledOpCode( "SWL" );
               // MemWriteCheck();

               break;

            case SW:           

               __asm {
                  movsx eax,Imm
                  mov   tValue32,eax
               };

               vAddr = (u32)(((s64)(tValue32)) + gpReg[rs]);

               __asm {
                  mov   eax,vAddr
                  and   eax,000000003h
                  cmp   eax,00h
                  jne   sw_addrexcept
               };

               *((s32 *)MemoryAccess(vAddr, MEMACCESS_WRITE)) = (s32)gpReg[rt];

               MemWriteCheck();

               break;

               sw_addrexcept:

                  wsprintf( szBuffer, "Address Error Exception at <0x%08X<.", PC );
                  MessageBox( hwndMain, szBuffer, 
                              "Store Word", MB_OK | MB_ICONEXCLAMATION );
                  break;

            case SDL:   

               ErrUnhandledOpCode( "SDL" );
               // MemWriteCheck();

               break;

            case SDR:     

               ErrUnhandledOpCode( "SDR" );
               // MemWriteCheck();

               break;

            case SWR:     

               ErrUnhandledOpCode( "SWR" );
               // MemWriteCheck();

               break;

            case CACHE:   

               // Just Ignored for Now.

               break;

            case LL:       

               ErrUnhandledOpCode( "LL" );

               break;

            case LWC1:     

               ErrUnhandledOpCode( "LWC1" );
/*
               __asm {
                  movsx eax,Imm
                  mov   tValue32,eax
               };

               vAddr = (UINT32)(((INT64)(tValue32)) + gpReg[rs]);

               __asm {
                  mov   eax,vAddr
                  and   eax,000000003h
                  cmp   eax,00h
                  jne   lwc1_addrexcept
               };

               cp1Reg[rt] = (INT64)(*( 
                  (INT32 *)PhysicalAddress( hwnd, vAddr, MEM_READ, MEM_WORD ) ));

               break;

               lwc1_addrexcept:

                  wsprintf( szBuffer, "Address Error Exception at <0x%08X>.", PC );
                  MessageBox( hwnd, szBuffer, "Load Word", MB_OK );

                  // TO DO: Address Error Exception
                  //        TLB Refill Exception
                  //        TLB Invalid Exception
                  //        Bus Error Exception
*/
                  break;

            case VLOAD:     

               // Get 'rd' and 'Funct' and 'sa' Subdivisions
               // ------------------------------------------
               // Bits [5:0]     Funct
               // Bits [10:6]    sa
               // Bits [15:11]   rd

               __asm {
                  mov   eax,[instrWord]   // Get 'rd' Value
                  mov   ecx,eax
                  shr   ecx,0Bh
                  and   ecx,1Fh
                  mov   [rd],cl
                  mov   edx,eax           // Get 'sa' Value
                  shr   edx,06h
                  and   edx,1Fh
                  mov   [sa],dl
                  and   eax,3Fh           // Get 'Funct' Value
                  mov   [Funct],al
               };

               switch( rd )
               {
                  case LBV:                  

                     ErrUnhandledOpCode( "LBV" );

                     break;
               
                  case LSV:

                     ErrUnhandledOpCode( "LSV" );

                     break;
               
                  case LLV:

                     ErrUnhandledOpCode( "LLV" );

                     break;
               
                  case LDV:

                     ErrUnhandledOpCode( "LDV" );

                     break;
               
                  case LQV:

                     ErrUnhandledOpCode( "LQV" );

                     break;
               
                  case LRV:

                     ErrUnhandledOpCode( "LRV" );

                     break;
               
                  case LPV:

                     ErrUnhandledOpCode( "LPV" );

                     break;
               
                  case LUV:

                     ErrUnhandledOpCode( "LUV" );

                     break;
               
                  case LHV:

                     ErrUnhandledOpCode( "LHV" );

                     break;
               
                  case LFV:

                     ErrUnhandledOpCode( "LFV" );

                     break;
               
                  case LWV:

                     ErrUnhandledOpCode( "LWV" );

                     break;
               
                  case LTV:

                     ErrUnhandledOpCode( "LTV" );

                     break;
               
                  default:

                     ErrUnhandledOpCode( "Vector Load" );

                     break;
               }

               break;

            case LLD:      

               ErrUnhandledOpCode( "LLD" );

               break;

            case LDC1:     

               ErrUnhandledOpCode( "LDC1" );

               break;

            case LDC2:     

               ErrUnhandledOpCode( "LDC2" );

               break;

            case LD:    

               //ErrUnhandledOpCode( "LD" );

               __asm {
                  movsx eax,Imm
                  mov   tValue32,eax
               };

               vAddr = (UINT32)(((INT64)(tValue32)) + gpReg[rs]);

               __asm {
                  mov   eax,vAddr
                  and   eax,000000003h
                  cmp   eax,00h
                  jne   ld_addrexcept
               };

               gpReg[rt] =  *(MemoryAccess(vAddr, MEMACCESS_READ));

               tValue64 = (s64)(*((s32 *)MemoryAccess(vAddr, MEMACCESS_READ)));
               tValue64 <<= 32;

               gpReg[rt] |= tValue64;

               break;

               ld_addrexcept:

                  wsprintf( szBuffer, "Address Error Exception at <0x%08X>.", PC );
                  MessageBox( hwndMain, szBuffer, "Load Doubleword", MB_OK );

                  // TO DO: Address Error Exception
                  //        TLB Refill Exception
                  //        TLB Invalid Exception
                  //        Bus Error Exception

                  break;

            case SC:     

               ErrUnhandledOpCode( "SC" );
               // MemWriteCheck();

               break;

            case SWC1:   

               ErrUnhandledOpCode( "SWC1" );
               // MemWriteCheck();

/*
               __asm {
                  movsx eax,Imm
                  mov   tValue32,eax
               };

               vAddr = (UINT32)(((INT64)(tValue32)) + gpReg[rs]);

               __asm {
                  mov   eax,vAddr
                  and   eax,000000003h
                  cmp   eax,00h
                  jne   swc1_addrexcept
               };

               *(
                  (INT32 *)PhysicalAddress( hwnd, vAddr, MEM_WRITE, MEM_WORD )
                ) = (INT32)(cp1Reg[rt]);

               break;

               swc1_addrexcept:

                  wsprintf( szBuffer, "Address Error Exception at <0x%08X<.", PC );
                  MessageBox( hwnd, szBuffer, "Store Word", MB_OK );

                  // TO DO: Address Error Exception
                  //        TLB Refill Exception
                  //        TLB Modification Exception
                  //        TLB Invalid Exception
                  //        Bus Error Exception
*/
                  break;

            case VSTORE:   

               // Get 'rd' and 'Funct' and 'sa' Subdivisions
               // ------------------------------------------
               // Bits [5:0]     Funct
               // Bits [10:6]    sa
               // Bits [15:11]   rd

               __asm {
                  mov   eax,[instrWord]   // Get 'rd' Value
                  mov   ecx,eax
                  shr   ecx,0Bh
                  and   ecx,1Fh
                  mov   [rd],cl
                  mov   edx,eax           // Get 'sa' Value
                  shr   edx,06h
                  and   edx,1Fh
                  mov   [sa],dl
                  and   eax,3Fh           // Get 'Funct' Value
                  mov   [Funct],al
               };

               switch( rd )
               {
                  case SBV:                  

                     ErrUnhandledOpCode( "SBV" );

                     break;
               
                  case SSV:

                     ErrUnhandledOpCode( "SSV" );

                     break;
               
                  case SLV:

                     ErrUnhandledOpCode( "SLV" );

                     break;
               
                  case SDV:

                     ErrUnhandledOpCode( "SDV" );

                     break;
               
                  case SQV:

                     ErrUnhandledOpCode( "SQV" );

                     break;
               
                  case SRV:

                     ErrUnhandledOpCode( "SRV" );

                     break;
               
                  case SPV:

                     ErrUnhandledOpCode( "SPV" );

                     break;
               
                  case SUV:

                     ErrUnhandledOpCode( "SUV" );

                     break;
               
                  case SHV:

                     ErrUnhandledOpCode( "SHV" );

                     break;
               
                  case SFV:

                     ErrUnhandledOpCode( "SFV" );

                     break;
               
                  case SWV:

                     ErrUnhandledOpCode( "SWV" );

                     break;
               
                  case STV:

                     ErrUnhandledOpCode( "STV" );

                     break;
               
                  default:

                     ErrUnhandledOpCode( "Vector Store" );
   
                     break;
               }

               break;

            case SCD:    

               ErrUnhandledOpCode( "SCD" );
               // MemWriteCheck();

               break;

            case SDC1:   

               ErrUnhandledOpCode( "SDC1" );
               // MemWriteCheck();
/*
               __asm {
                  movsx eax,Imm
                  mov   tValue32,eax
               };

               vAddr = (UINT32)(((INT64)(tValue32)) + gpReg[rs]);

               __asm {
                  mov   eax,vAddr
                  and   eax,000000003h
                  cmp   eax,00h
                  jne   sdc1_addrexcept
               };

               *(
                  (INT32 *)PhysicalAddress( hwnd, vAddr + 4, MEM_WRITE, MEM_WORD )
                ) = (INT32)(cp1Reg[rt] & 0x00000000FFFFFFFF);
               *(
                  (INT32 *)PhysicalAddress( hwnd, vAddr, MEM_WRITE, MEM_WORD )
                ) = (INT32)((cp1Reg[rt] & 0xFFFFFFFF00000000) >> 16);

               break;

               sdc1_addrexcept:

                  wsprintf( szBuffer, "Address Error Exception at <0x%08X>.", PC );
                  MessageBox( hwnd, szBuffer, "Store Doubleword from CP1", MB_OK );

                  // TO DO: Address Error Exception
                  //        TLB Refill Exception
                  //        TLB Invalid Exception
                  //        Bus Error Exception
*/
                  break;

            case SDC2:   

               ErrUnhandledOpCode( "SDC2" );
               // MemWriteCheck();

               break;

            case SD:     

               ErrUnhandledOpCode( "SD" );
               // MemWriteCheck();
/*
               __asm {
                  movsx eax,Imm
                  mov   tValue32,eax
               };

               vAddr = (UINT32)(((INT64)(tValue32)) + gpReg[rs]);

               __asm {
                  mov   eax,vAddr
                  and   eax,000000003h
                  cmp   eax,00h
                  jne   sd_addrexcept
               };

               *(
                  (INT32 *)PhysicalAddress( hwnd, vAddr + 4, MEM_WRITE, MEM_WORD )
                ) = (INT32)(gpReg[rt] & 0x00000000FFFFFFFF);
               *(
                  (INT32 *)PhysicalAddress( hwnd, vAddr, MEM_WRITE, MEM_WORD )
                ) = (INT32)((gpReg[rt] & 0xFFFFFFFF00000000) >> 16);

               break;

               sd_addrexcept:

                  wsprintf( szBuffer, "Address Error Exception at <0x%08X>.", PC );
                  MessageBox( hwnd, szBuffer, "Store Doubleword", MB_OK );

                  // TO DO: Address Error Exception
                  //        TLB Refill Exception
                  //        TLB Invalid Exception
                  //        Bus Error Exception
*/
                  break;

            default:
   
               // Main Processor Unhandled OpCode

               ErrUnhandledOpCode( "CPU" );

               break;
         };

         // Increment Count Register

         cp0Reg[C0_COUNT]++;

         // Increment Program Counter or Process a BranchDelay

         switch( branchDelay )
         {
            // Next Instruction is the Branch Address

            case 2: 

               branchDelay = 0;
               PC = branchPC;

               break;
         
            // A Branch Address was Stored in the Branch Program Counter.
            // The Next Instruction is a PC + 4 (Case 0 is Executed).
            // The Next Instruction is at the Branch Program Counter

            case  1:

               branchDelay = 2;

            // Normal Execution

            case 0:

               PC += 4;

               break;
         } 
      }

      return( 0 );
   }


////////////////////////////////////////////////////////////////////////////////
// Hardware Reset - Set Registers, Status Bits etc to Initial State.

   void InitHardware( void )
   {
      // Initialise Random Register

      cp0Reg[C0_RAND] = 0x000000000000001F;

      // Initialise the Status Register - This is a read/write that contains
      // the operating mode, interrupt enabling and diagnostic states of the
      // processor.
      //
      // 31                             0
      // xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
      // 0111 (CU)                        Coprocessors 0,1,2 Usable,3 Unsable
      //     0 (RP)                       Reduced Power Mode, 0 = Full Speed
      //      1 (FR)                      Floating Point Regs, 1 = 32 Regs
      //       0 (RE)                     Reverse-Endian bit, 0 = Disabled
      //        000000000 (DS)            Diagnostic Status Field
      //                 00000000 (IM)    Interrupt Mask Field, 1 = Enabled
      //                    (KX) 0        64 Bit Addrsng Kernel Mode, 0 = 32Bit
      //                     (SX) 0       64 Bit Addr Supervisor Mode, 0 = 32Bit
      //                      (UX) 0      64 Bit Addressing User Mode, 0 = 32Bit
      //                      (KSU) 10    Indicates Operating Mode, 10 = User
      //                        (ERL) 0   Indicates Error Level, 0 = Normal
      //                         (EXL) 0  Exception Level, 0 = Normal
      //                           (IE) 1 Global Interrupt Enable, 1 = Enable
        
      cp0Reg[C0_SR] = 0x0000000070400004;

      // Initialise the Cause Register - This is a read/write register that
      // holds the cause of the exception that has occurred last.
      //
      // 31                             0
      // xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
      // 0 (BD)                           Last exception executed in delay slot?
      //  0                               Reserved for Future Use
      //   00 (CE)                        Coproc. No. of last Unusable Exception
      //     000000000000                 Reserved for Future Use
      //                 00000000 (IP)    Indicates an Interrupt is Pending
      //                         0        Reserved for Future Use
      //                (ExcCode) 00000   Exception Code Field
      //                               00 Reserved for Future Use

      cp0Reg[C0_CAUSE] = 0x0000000000000000;

      // Initialise the Exception Program Counter (EPC) Register
      // This register contains the address at which processing 
      // resumes after an exception has been serviced.

      cp0Reg[C0_EPC] = 0x0000000000000000;

      // Initialise Processor Revision ID Register

      cp0Reg[C0_PRID] = 0x0000000000000411;

      // Initialise the Config Register
     
      cp0Reg[C0_CONFIG] = 0x0000000000066463;

      // Initialise the Error Exception Program Counter (Error EPC) Register
      
      cp0Reg[C0_ERROR_EPC] = 0x0000000000000000;

      // Initialise Program Counter Register to Point at Boot Code

      PC = baseAddress;

      // Initialise HI and LO Registers

      HI = 0;
      LO = 0;

      // Initialise General Purpose Register 0 (r0) to 0
      // This is hardwired in the real machine.

      gpReg[0] = 0x0000000000000000;

      // Initialise other General Purpose Registers

      gpReg[0]  = 0x0000000000000000;
      gpReg[20] = 0x0000000000000001;
      gpReg[22] = 0x000000000000003F;
      gpReg[29] = 0x0000000000400000;

      // Initialise Floating Point Implementation/Revision Register

      FCR0 = 0x00000B11;

      // Initialise MIPS Version Register (0x04300004)

      *(miVersionReg) = 0x01010101;

      InitInterrupts();
   }


////////////////////////////////////////////////////////////////////////////////
// Unhandle OpCode Error Message

   void ErrUnhandledOpCode( LPSTR desc )
   {
      wsprintf( szBuffer, 
         "Error: Unhandled Opcode %s at 0x%08X "
         "<%0.2X %0.2X %0.2X %0.2X>\n\nAction: Terminating Execution",
         desc, PC, ((instrWord & 0xFF000000) >> 24),
                   ((instrWord & 0x00FF0000) >> 16),
                   ((instrWord & 0x0000FF00) >>  8),
                   (instrWord & 0x000000FF) );

      MessageBox( hwndMain, szBuffer, 
                  "Unhandled Opcode", 
                  MB_OK | MB_ICONEXCLAMATION );

      cpuIsRunning = FALSE;
   }
