////////////////////////////////////////////////////////////////////////////////
// r4300.h
// Core R4300 Definitions
//
// Nintendo, Nintendo 64, or N64 are the Copyright Nintendo of America, Inc. 
// and Nintendo of Japan, Inc. Nintendo 64, N64, and the Nintendo 64 Logo are 
// Registered Trademarks of Nintendo of America, Inc. and Nintendo of Japan, 
// Inc. The NSFE project is in no way linked to Nintendo.
//
// All other Trademarks and Copyrights are Acknowledged.
//
// This software is provided "as is" without any expressed or implied warranty.
// -----------------------------------------------------------------------------
// Update History, r4300.h
//
// 00001 16-02-1998 "Initial Code Release.                                     "


#ifndef _R4300_H_
#define _R4300_H_


////////////////////////////////////////////////////////////////////////////////
// Include Files

#ifndef _TYPEDEFS_H_
#include "typedefs.h"                  // Core Type Definitions
#endif // _TYPEDEFS_H_


////////////////////////////////////////////////////////////////////////////////
// Segment Base Addresses and Sizes
 
#define  KUBASE            0x00000000
#define  KUSIZE		      0x80000000
#define  K0BASE		      0x80000000
#define  K0SIZE		      0x20000000
#define  K1BASE		      0xA0000000
#define  K1SIZE		      0x20000000
#define  K2BASE		      0xC0000000
#define  K2SIZE		      0x20000000


////////////////////////////////////////////////////////////////////////////////
// Address Conversion Macros
 
#define  K0_TO_K1(x)       ((u32)(x)|0xA0000000)   // KSEG0 to KSEG1
#define  K1_TO_K0(x)       ((u32)(x)&0x9FFFFFFF)   // KSEG1 to KSEG0
#define  K0_TO_PHYS(x)     ((u32)(x)&0x1FFFFFFF)   // KSEG0 to Physical
#define  K1_TO_PHYS(x)     ((u32)(x)&0x1FFFFFFF)   // KSEG1 to Physical
#define  KDM_TO_PHYS(x)    ((u32)(x)&0x1FFFFFFF)   // Direct Mapped to Physical
#define  PHYS_TO_K0(x)     ((u32)(x)|0x80000000)   // Physical to KSEG0
#define  PHYS_TO_K1(x)     ((u32)(x)|0xA0000000)   // Physical to KSEG1


////////////////////////////////////////////////////////////////////////////////
// Address Predicates
 
#define  IS_KSEG0(x)       ((u32)(x) >= K0BASE && (u32)(x) < K1BASE)
#define  IS_KSEG1(x)       ((u32)(x) >= K1BASE && (u32)(x) < K2BASE)
#define  IS_KSEGDM(x)      ((u32)(x) >= K0BASE && (u32)(x) < K2BASE)
#define  IS_KSEG2(x)       ((u32)(x) >= K2BASE && (u32)(x) < KPTE_SHDUBASE)
#define  IS_KPTESEG(x)     ((u32)(x) >= KPTE_SHDUBASE)
#define  IS_KUSEG(x)       ((u32)(x) < K0BASE)

#define  IS_RDRAMREGS(x)   ((u32)(x) >= 0x03F00000 && (u32)(x) < 0x03F80027)
#define  IS_SPMEM(x)       ((u32)(x) >= 0x04000000 && (u32)(x) < 0x04002000)


////////////////////////////////////////////////////////////////////////////////
// Memory Access Flags

#define  MEMACCESS_READ    0x00000001
#define  MEMACCESS_WRITE   0x00000002


////////////////////////////////////////////////////////////////////////////////
// Exception Vectors

#define  SIZE_EXCVEC	      0x80			         // Size of an Exception Vector
#define  UT_VEC		      K0BASE			      // UTLBmiss Vector
#define  R_VEC		         (K1BASE+0x1FC00000)	// Reset Vector
#define  XUT_VEC		      (K0BASE+0x80)		   // Extended Address TLBmiss 
#define  ECC_VEC		      (K0BASE+0x100)		   // ECC Exception Vector
#define  E_VEC		         (K0BASE+0x180)		   // General Exception Vector 


////////////////////////////////////////////////////////////////////////////////
// TLB Size Constants

#define  NTLBENTRIES       31	

#define  TLBHI_VPN2MASK    0xFFFFE000
#define  TLBHI_VPN2SHIFT   13
#define  TLBHI_PIDMASK     0xFF
#define  TLBHI_PIDSHIFT    0
#define  TLBHI_NPID        255

#define  TLBLO_PFNMASK     0x3FFFFFC0
#define  TLBLO_PFNSHIFT    6
#define  TLBLO_CACHMASK    0x38		            // Cache Coherency Algorithm
#define  TLBLO_CACHSHIFT   3
#define  TLBLO_UNCACHED		0x10                 // Not Cached
#define  TLBLO_NONCOHRNT   0x18		            // Cacheable Non-Coherent
#define  TLBLO_EXLWR       0x28                 // Exclusive Write
#define  TLBLO_D           0x4                  // Writeable 
#define  TLBLO_V           0x2                  // Valid Bit
#define  TLBLO_G           0x1                  // Global Access Bit

#define  TLBINX_PROBE      0x80000000
#define  TLBINX_INXMASK    0x3F
#define  TLBINX_INXSHIFT   0

#define  TLBRAND_RANDMASK  0x3F
#define  TLBRAND_RANDSHIFT 0

#define  TLBWIRED_WIREDMASK 0x3F

#define  TLBCTXT_BASEMASK  0xFF800000
#define  TLBCTXT_BASESHIFT 23
#define  TLBCTXT_BASEBITS  9

#define  TLBCTXT_VPNMASK   0x7FFFF0
#define  TLBCTXT_VPNSHIFT  4

#define  TLBPGMASK_4K      0x0
#define  TLBPGMASK_16K     0x6000
#define  TLBPGMASK_64K     0x1E000


////////////////////////////////////////////////////////////////////////////////
// Status Register
 
#define  SR_CUMASK         0xF0000000     // Coprocessor Usable Bits
#define  SR_CU3            0x80000000     // Coprocessor 3 usable
#define  SR_CU2            0x40000000     // Coprocessor 2 usable
#define  SR_CU1            0x20000000     // Coprocessor 1 usable
#define  SR_CU0            0x10000000     // Coprocessor 0 usable
#define  SR_RP             0x08000000     // Reduced power (quarter speed)
#define  SR_FR             0x04000000     // MIPS III FP register mode
#define  SR_RE             0x02000000     // Reverse endian
#define  SR_ITS            0x01000000     // Instruction trace support
#define  SR_BEV            0x00400000     // Use boot exception vectors
#define  SR_TS             0x00200000     // TLB shutdown
#define  SR_SR             0x00100000     // Soft reset occured
#define  SR_CH             0x00040000     // Cache hit for last 'cache' op
#define  SR_CE             0x00020000     // Create ECC
#define  SR_DE             0x00010000     // ECC of parity does not cause error

// Interrupt Enable Bits

#define  SR_IMASK          0x0000FF00     // Interrupt Mask
#define  SR_IMASK8         0x00000000     // Mask Level 8
#define  SR_IMASK7         0x00008000     // Mask Level 7
#define  SR_IMASK6         0x0000C000     // Mask Level 6
#define  SR_IMASK5         0x0000E000     // Mask Level 5
#define  SR_IMASK4         0x0000F000     // Mask Level 4
#define  SR_IMASK3         0x0000F800     // Mask Level 3
#define  SR_IMASK2         0x0000FC00     // Mask Level 2
#define  SR_IMASK1         0x0000FE00     // Mask Level 1
#define  SR_IMASK0         0x0000FF00     // Mask Level 0

#define  SR_IBIT8          0x00008000     // Bit Level 8
#define  SR_IBIT7          0x00004000     // Bit Level 7
#define  SR_IBIT6          0x00002000     // Bit Level 6
#define  SR_IBIT5          0x00001000     // Bit Level 5
#define  SR_IBIT4          0x00000800     // Bit Level 4
#define  SR_IBIT3          0x00000400     // Bit Level 3
#define  SR_IBIT2          0x00000200     // Bit Level 2
#define  SR_IBIT1          0x00000100     // Bit Level 1

#define  SR_IMASKSHIFT     8

#define  SR_KX             0x00000080     // Extended-Addr TLB Vec in Kernel
#define  SR_SX             0x00000040     // Extended-Addr TLB Vec Supervisor
#define  SR_UX             0x00000020     // Extended-Addr TLB Vec in User Mode
#define  SR_KSU_MASK       0x00000018     // Mode Mask
#define  SR_KSU_USR        0x00000010     // User Mode
#define  SR_KSU_SUP        0x00000008     // Supervisor Mode
#define  SR_KSU_KER        0x00000000     // Kernel Mode
#define  SR_ERL            0x00000004     // Error Level, 1 => Cache Error
#define  SR_EXL            0x00000002     // Exception Level, 1 => Exception
#define  SR_IE             0x00000001     // Interrupt Enable, 1 => Enable


////////////////////////////////////////////////////////////////////////////////
// Cause Register
 
#define  CAUSE_BD          0x80000000     // Branch Delay Slot
#define  CAUSE_CEMASK      0x30000000     // Coprocessor Error
#define  CAUSE_CESHIFT     28

// Interrupt Pending Bits

#define  CAUSE_IP8         0x00008000     // External Level 8 Pending - COMPARE
#define  CAUSE_IP7         0x00004000     // External Level 7 Pending - INT4
#define  CAUSE_IP6         0x00002000     // External Level 6 Pending - INT3
#define  CAUSE_IP5         0x00001000     // External Level 5 Pending - INT2
#define  CAUSE_IP4         0x00000800     // External Level 4 Pending - INT1
#define  CAUSE_IP3         0x00000400     // External Level 3 Pending - INT0
#define  CAUSE_SW2         0x00000200     // Software Level 2 Pending
#define  CAUSE_SW1         0x00000100     // Software Level 1 Pending

#define  CAUSE_IPMASK      0x0000FF00     // Pending Interrupt Mask
#define  CAUSE_IP01MASK    0x00000300     // IP0 and IP1 Pending Interrupt Mask
#define  CAUSE_IPSHIFT     8

#define  CAUSE_EXCMASK     0x0000007C     // Cause Code Bits

#define  CAUSE_EXCSHIFT    2

// Cause Register Exception Codes

#define  EXC_CODE(x)       ((x)<<2)

// Hardware Exception Codes

#define  EXC_INT           EXC_CODE(0)    // Interrupt
#define  EXC_MOD           EXC_CODE(1)    // TLB Modification
#define  EXC_RMISS         EXC_CODE(2)    // Read TLB Miss
#define  EXC_WMISS         EXC_CODE(3)    // Write TLB Miss
#define  EXC_RADE          EXC_CODE(4)    // Read Address Error
#define  EXC_WADE          EXC_CODE(5)    // Write Address Error
#define  EXC_IBE           EXC_CODE(6)    // Instruction Bus Error
#define  EXC_DBE           EXC_CODE(7)    // Data Bus Error
#define  EXC_SYSCALL       EXC_CODE(8)    // SYSCALL
#define  EXC_BREAK         EXC_CODE(9)    // BREAKPOINT
#define  EXC_II            EXC_CODE(10)   // Illegal Instruction
#define  EXC_CPU           EXC_CODE(11)   // Coprocessor Unusable 
#define  EXC_OV            EXC_CODE(12)   // Overflow
#define  EXC_TRAP          EXC_CODE(13)   // Trap exception
#define  EXC_VCEI          EXC_CODE(14)   // Virt. Coherency on Inst. Fetch
#define  EXC_FPE           EXC_CODE(15)   // Floating Point Exception
#define  EXC_WATCH         EXC_CODE(23)   // Watchpoint reference
#define  EXC_VCED          EXC_CODE(31)   // Virt. Coherency on Data Read


////////////////////////////////////////////////////////////////////////////////
// Cache Instruction Definitions

// Target Cache

#define  CACH_PI           0x0      // Specifies Primary Instr. Cache
#define  CACH_PD           0x1      // Primary Data Cache
#define  CACH_SI           0x2      // Secondary Instruction Cache
#define  CACH_SD           0x3      // Secondary Data Cache

// Cache Operations

#define  C_IINV            0x0      // Index Invalidate (Inst, 2nd Inst)
#define  C_IWBINV          0x0      // Index Writeback Inval (d, sd)
#define  C_ILT             0x4      // Index Load Tag (All)
#define  C_IST             0x8      // Index Store Tag (All)
#define  C_CDX             0xC      // Create Dirty Exclusive (d, sd)
#define  C_HINV            0x10     // Hit Invalidate (all)
#define  C_HWBINV          0x14     // Hit Writeback Inv. (d, sd)
#define  C_FILL            0x14     // Fill (I)
#define  C_HWB             0x18     // Hit Writeback (i, d, sd)
#define  C_HSV             0x1C     // Hit Set Virt. (si, sd)


////////////////////////////////////////////////////////////////////////////////
// Cache Size Definitions
 
#define  ICACHE_SIZE       0x4000         // 16K
#define  ICACHE_LINESIZE   32	            // 8 Words
#define  ICACHE_LINEMASK   (ICACHE_LINESIZE-1)

#define  DCACHE_SIZE       0x2000         // 8K 
#define  DCACHE_LINESIZE   16             // 4 Words
#define  DCACHE_LINEMASK   (DCACHE_LINESIZE-1)


////////////////////////////////////////////////////////////////////////////////
// Coprocessor 0

// Coprocessor 0 Operations
 
#define  C0_READI          0x1      // Read ITLB Entry Addressed by C0_INDEX
#define  C0_WRITEI         0x2      // Write ITLB Entry Addressed by C0_INDEX
#define  C0_WRITER         0x6      // Write ITLB Entry Addressed by C0_RAND
#define  C0_PROBE          0x8      // Probe for ITLB Entry Addressed by TLBHI
#define  C0_RFE            0x10     // Restore for Exception

// C0_PRID Defines

#define  C0_IMPMASK        0xFF00
#define  C0_IMPSHIFT       8
#define  C0_REVMASK        0xFF
#define  C0_MAJREVMASK     0xF0
#define  C0_MAJREVSHIFT    4
#define  C0_MINREVMASK     0xF

// C0_CONFIG Register Definitions
 
#define  CONFIG_CM         0x80000000  // 1 == Master-Checker Enabled
#define  CONFIG_EC         0x70000000  // System Clock Ratio
#define  CONFIG_EC_1_1     0x6         // System Clock Ratio 1:1
#define  CONFIG_EC_3_2     0x7         // System Clock Ratio 1.5:1
#define  CONFIG_EC_2_1     0x0         // System Clock Ratio 2:1
#define  CONFIG_EC_3_1     0x1         // System Clock Ratio 3:1 
#define  CONFIG_EP         0x0F000000  // Transmit Data Pattern
#define  CONFIG_SB         0x00C00000  // Secondary Cache Block Size

#define  CONFIG_SS         0x00200000  // Split SCache: 0 == I&D Combined
#define  CONFIG_SW         0x00100000  // SCache port: 0==128, 1==64
#define  CONFIG_EW         0x000C0000  // System Port Width: 0==64, 1==32
#define  CONFIG_SC         0x00020000  // 0 -> 2nd Cache Present
#define  CONFIG_SM         0x00010000  // 0 -> Dirty Shared Coherency Enabled
#define  CONFIG_BE         0x00008000  // Endianess: 1 --> BE
#define  CONFIG_EM         0x00004000  // 1 -> ECC Mode, 0 -> Parity
#define  CONFIG_EB         0x00002000  // Block Order:1->Sequent,0->Subblock

#define  CONFIG_IC         0x00000E00  // Primary Icache Size
#define  CONFIG_DC         0x000001C0  // Primary Dcache Size
#define  CONFIG_IB         0x00000020  // Icache Block Size
#define  CONFIG_DB         0x00000010  // Dcache Block Size
#define  CONFIG_CU         0x00000008  // Update on Store-Conditional
#define  CONFIG_K0         0x00000007  // K0SEG Coherency Algorithm

#define  CONFIG_UNCACHED      0x00000002  // K0 is Uncached
#define  CONFIG_NONCOHRNT     0x00000003
#define  CONFIG_COHRNT_EXLWR  0x00000005
#define  CONFIG_SB_SHFT       22          // Shift SB to Bit Position 0
#define  CONFIG_IC_SHFT       9           // Shift IC to Bit Position 0
#define  CONFIG_DC_SHFT       6           // Shift DC to Bit Position 0
#define  CONFIG_BE_SHFT       15          // Shift BE to Bit Position 0

// C0_TAGLO Definitions for Setting/Getting Cache States and Physaddr Bits
 
#define  SADDRMASK         0xFFFFE000  // 31..13 -> Scache Paddr Bits 35..17
#define  SVINDEXMASK       0x00000380  // 9..7: Prim Virt Index Bits 14..12
#define  SSTATEMASK        0x00001C00  // Bits 12..10 Hold Scache Line State
#define  SINVALID          0x00000000  // Invalid --> 000 == state 0
#define  SCLEANEXCL        0x00001000  // Clean Exclusive --> 100 == state 4
#define  SDIRTYEXCL        0x00001400  // Dirty Exclusive --> 101 == state 5
#define  SECC_MASK         0x0000007F  // Low 7 Bits are ECC for the Tag
#define  SADDR_SHIFT       4           // Shift STagLo (31..13) to 35..17

#define  PADDRMASK         0xFFFFFF00  // PTagLo31..8->Prim Paddr Bits 35..12
#define  PADDR_SHIFT       4           // Roll Bits 35..12 Down to 31..8
#define  PSTATEMASK        0x00C0      // Bits 7..6 Hold Primary Line State
#define  PINVALID          0x0000      // Invalid --> 000 == state 0
#define  PCLEANEXCL        0x0080      // Clean Exclusive --> 10 == state 2
#define  PDIRTYEXCL        0x00C0      // Dirty Exclusive --> 11 == state 3
#define  PPARITY_MASK      0x0001      // Low Bit is Parity Bit (even).

// C0_CACHE_ERR Definitions.

#define  CACHERR_ER           0x80000000  // 0: Inst Ref, 1: Data Ref
#define  CACHERR_EC           0x40000000  // 0: Primary, 1: Secondary
#define  CACHERR_ED           0x20000000  // 1: Data Error
#define  CACHERR_ET           0x10000000  // 1: Tag Error
#define  CACHERR_ES           0x08000000  // 1: External Ref
#define  CACHERR_EE           0x04000000  // Error on SysAD bus
#define  CACHERR_EB           0x02000000  
#define  CACHERR_EI           0x01000000  
#define  CACHERR_SIDX_MASK    0x003FFFF8  // Secondary Cache Index
#define  CACHERR_PIDX_MASK    0x00000007  // Primary Cache Index
#define  CACHERR_PIDX_SHIFT   12          // Bits 2..0 are Paddr14..12

// Coprocessor 0 Registers
 
#define  C0_INX            0
#define  C0_RAND           1
#define  C0_ENTRYLO0       2
#define  C0_ENTRYLO1       3
#define  C0_CONTEXT        4
#define  C0_PAGEMASK       5           // Page Mask
#define  C0_WIRED          6           // Wired Entries in TLB
#define  C0_BADVADDR       8
#define  C0_COUNT          9           // Free-Running Counter
#define  C0_ENTRYHI        10
#define  C0_SR             12
#define  C0_CAUSE          13
#define  C0_EPC            14
#define  C0_PRID           15          // Revision Identifier
#define  C0_COMPARE        11          // Counter Comparison Reg.
#define  C0_CONFIG         16          // Hardware Configuration 
#define  C0_LLADDR         17          // Load Linked Address
#define  C0_WATCHLO        18          // Watchpoint
#define  C0_WATCHHI        19          // Watchpoint
#define  C0_ECC            26          // S-Cache ECC and Primary Parity
#define  C0_CACHE_ERR      27          // Cache Error Status
#define  C0_TAGLO          28          // Cache Operations
#define  C0_TAGHI          29          // Cache Operations
#define  C0_ERROR_EPC      30          // ECC Error Prg. Counter


////////////////////////////////////////////////////////////////////////////////
// Floating Point Status Register 
 
#define  FPCSR_FS          0x01000000  // Flush Denorm to Zero 
#define  FPCSR_C           0x00800000  // Condition Bit
#define  FPCSR_CE          0x00020000  // Cause: Unimplemented Operation
#define  FPCSR_CV          0x00010000  // Cause: Invalid Operation
#define  FPCSR_CZ          0x00008000  // Cause: Division by Zero 
#define  FPCSR_CO          0x00004000  // Cause: Overflow
#define  FPCSR_CU          0x00002000  // Cause: Underflow
#define  FPCSR_CI          0x00001000  // Cause: Inexact Operation
#define  FPCSR_EV          0x00000800  // Enable: Invalid Operation
#define  FPCSR_EZ          0x00000400  // Enable: Division by Zero
#define  FPCSR_EO          0x00000200  // Enable: Overflow
#define  FPCSR_EU          0x00000100  // Enable: Underflow
#define  FPCSR_EI          0x00000080  // Enable: Inexact Operation
#define  FPCSR_FV          0x00000040  // Flag: Invalid Operation
#define  FPCSR_FZ          0x00000020  // Flag: Division by Zero
#define  FPCSR_FO          0x00000010  // Flag: Overflow
#define  FPCSR_FU          0x00000008  // Flag: Underflow
#define  FPCSR_FI          0x00000004  // Flag: Inexact Operation
#define  FPCSR_RM_MASK     0x00000003  // Rounding Mode Mask
#define  FPCSR_RM_RN       0x00000000  // Round to Nearest
#define  FPCSR_RM_RZ       0x00000001  // Round to Zero
#define  FPCSR_RM_RP       0x00000002  // Round to Positive Infinity
#define  FPCSR_RM_RM       0x00000003  // Round to Negative Infinity 


////////////////////////////////////////////////////////////////////////////////
// CPU/Memory Action Flags

#define  CLEAR_ACTIONS     0x00000000  // Clear CPU Action Flags
#define  CPU_INTERRUPTS    0x00000001  // Check the Core CPU Interrupts
#define  MI_INTERRUPTS     0x00000002  // Check MI Interrupts
#define  UPDATE_MI_REG     0x00000004  // Update MI Registers
#define  CHECK_VI_REG      0x00000008  // Check VI Registers
#define  UPDATE_PI_REG     0x00000010  // Update PI Registers

// DMA Transfers (The 8 MSBs Define Various DMA Transfers)

#define  DMA_REQUEST_MASK  0xFF000000  // Mask DMA Requests
#define  PI_DMA_WRITE      0x80000000  // PI DMA Write Request
#define  PIF2DRAM_DMA      0x40000000  // PIF to DRAM DMA Request
#define  DRAM2PIF_DMA      0x20000000  // DRAM to PIF DMA Request


#endif // _R4300_H_
