////////////////////////////////////////////////////////////////////////////////
// rcpmem.h
// Register and Bit Definitions for the RCP Memory Map
//
// Nintendo, Nintendo 64, or N64 are the Copyright Nintendo of America, Inc. 
// and Nintendo of Japan, Inc. Nintendo 64, N64, and the Nintendo 64 Logo are 
// Registered Trademarks of Nintendo of America, Inc. and Nintendo of Japan, 
// Inc. The NSFE project is in no way linked to Nintendo.
//
// All other Trademarks and Copyrights are Acknowledged.
//
// This software is provided "as is" without any expressed or implied warranty.
// -----------------------------------------------------------------------------
// Update History, rcpmem.h
//
// 00001 16-02-1998 "Initial Code Release.                                     "


#ifndef _RCPMEM_H_
#define _RCPMEM_H_


////////////////////////////////////////////////////////////////////////////////
// File Includes

#ifndef _TYPEDEFS_H_
#include "typedefs.h"                  // Core Type Definitions
#endif // _TYPEDEFS_H_


////////////////////////////////////////////////////////////////////////////////
// RDRAM Memory (Assumes that Maximum Size is 4 MB)
 
#define  RDRAM_0_START           0x00000000
#define  RDRAM_0_END             0x001FFFFF
#define  RDRAM_1_START           0x00200000
#define  RDRAM_1_END             0x0040FFFF

#define  RDRAM_START             RDRAM_0_START
#define  RDRAM_END               RDRAM_1_END

// Address Predicates

#define  IS_RDRAM(x)             ((u32)(x)>=RDRAM_START && (u32)(x)<RDRAM_END)


////////////////////////////////////////////////////////////////////////////////
// RDRAM Registers (0x03F00000 - 0x03FFFFFF)
 
#define  RDRAM_CONFIG_REG        0x03F00000
#define  RDRAM_DEVICE_TYPE_REG   0x03F00000
#define  RDRAM_DEVICE_ID_REG     0x03F00004
#define  RDRAM_DELAY_REG         0x03F00008
#define  RDRAM_MODE_REG          0x03F0000C
#define  RDRAM_REF_INTERVAL_REG  0x03F00010
#define  RDRAM_REF_ROW_REG       0x03F00014
#define  RDRAM_RAS_INTERVAL_REG  0x03F00018
#define  RDRAM_MIN_INTERVAL_REG  0x03F0001C
#define  RDRAM_ADDR_SELECT_REG   0x03F00020
#define  RDRAM_DEVICE_MANUF_REG  0x03F00024

#define  RDRAM_0_DEVICE_ID       0
#define  RDRAM_1_DEVICE_ID       1

#define  RDRAM_RESET_MODE        0
#define  RDRAM_ACTIVE_MODE       1
#define  RDRAM_STANDBY_MODE      2

#define  RDRAM_LENGTH            0x00400000
#define  RDRAM_0_BASE_ADDRESS    0x00000000
#define  RDRAM_1_BASE_ADDRESS    0x00200000

#define  RDRAM_0_CONFIG          0x00000
#define  RDRAM_1_CONFIG          0x00400
#define  RDRAM_GLOBAL_CONFIG     0x80000


////////////////////////////////////////////////////////////////////////////////
// PIF Memory

#define  PIF_ROM_START           0x1FC00000
#define  PIF_ROM_END             0x1FC007BF
#define  PIF_RAM_START           0x1FC007C0
#define  PIF_RAM_END             0x1FC007FF

// Controller Channel 
// Each Game Controller Channel has 4 Error Bits that are Defined in Bit 6-7 of
// the Rx and Tx Data Size Area Bytes. Programmers need to Clear these Bits
// when Setting the Tx/Rx Size Area Values for a Channel
 
#define  CHNL_ERR_NORESP         0x80     // Bit 7 (Rx): No Response Error
#define  CHNL_ERR_OVERRUN        0x40     // Bit 6 (Rx): Overrun Error
#define  CHNL_ERR_FRAME          0x80     // Bit 7 (Tx): Frame Error
#define  CHNL_ERR_COLLISION      0x40     // Bit 6 (Tx): Collision Error
#define  CHNL_ERR_MASK           0xC0     // Bit 6-7: Channel Errors


////////////////////////////////////////////////////////////////////////////////
// External Device Information
 
#define  DEVICE_TYPE_CART        0        // ROM Cartridge
#define  DEVICE_TYPE_BULK        1        // ROM Bulk
#define  DEVICE_TYPE_64DD        2        // 64 Disk Drive


////////////////////////////////////////////////////////////////////////////////
// SP Memory

#define  SP_DMEM_START           0x04000000  // Read/Write
#define  SP_DMEM_END             0x04000FFF
#define  SP_IMEM_START           0x04001000	// Read/Write
#define  SP_IMEM_END             0x04001FFF


////////////////////////////////////////////////////////////////////////////////
// SP CP0 Registers
 
// SP Memory Address
// READ/WRITE
// Bits [11:0]    DMEM/IMEM Address
// Bits [12]      0=DMEM, 1=IMEM 

#define  SP_MEM_ADDR_REG         0x04040000     // Master
						
// SP DRAM DMA Address
// READ/WRITE
// Bits [23:0]    RDRAM Address

#define  SP_DRAM_ADDR_REG        0x04040004     // Slave

// SP Read DMA Length (Direction: I/DMEM <- RDRAM)
// READ/WRITE
// Bits [11:0]    Length
// Bits [19:12]   Count
// Bits [31:20]   Skip

#define  SP_RD_LEN_REG           0x04040008

// SP Write DMA Length (Direction: I/DMEM -> RDRAM)
// READ/WRITE
// Bits [11:0]    Length
// Bits [19:12]   Count
// Bits [31:20]   Skip

#define  SP_WR_LEN_REG           0x0404000C

// SP Status
// READ/WRITE
// Bits [14:0]    Valid Bits (See below for Write/Read Mode)

#define  SP_STATUS_REG           0x04040010

// SP DMA Full
// READ
// Bits [0]       Valid Bit - DMA Full

#define  SP_DMA_FULL_REG         0x04040014

// SP DMA Busy
// READ
// Bits [0]       Valid Bit - DMA Busy 

#define  SP_DMA_BUSY_REG         0x04040018

// SP semaphore
// READ
// Bits [0]       Semaphore Flag (Set on Read)
// WRITE
// Bits []        Clear Semaphore Flag

#define  SP_SEMAPHORE_REG        0x0404001C

// SP PC
// READ/WRITE
// Bits [11:0]    Program Counter

#define  SP_PC_REG               0x04080000

// SP MEM Address
// Bits [12]      Specifies if Address is IMEM (1) or DMEM (0)

#define  SP_DMA_DMEM             0x0000
#define  SP_DMA_IMEM             0x1000

// Values to Clear/Set bit in Status Register (SP_STATUS_REG - Write)
 
#define  SP_CLR_HALT             0x00001        // Bit  0: Clear Halt
#define  SP_SET_HALT             0x00002        // Bit  1: Set Halt
#define  SP_CLR_BROKE            0x00004        // Bit  2: Clear Broke
#define  SP_CLR_INTR             0x00008        // Bit  3: Clear Intr
#define  SP_SET_INTR             0x00010        // Bit  4: Set Intr
#define  SP_CLR_SSTEP            0x00020        // Bit  5: Clear Sste
#define  SP_SET_SSTEP            0x00040        // Bit  6: Set Sstep
#define  SP_CLR_INTR_BREAK       0x00080        // Bit  7: Clear Intr on Break
#define  SP_SET_INTR_BREAK       0x00100        // Bit  8: Set Intr on Break
#define  SP_CLR_SIG0             0x00200        // Bit  9: Clear Signal 0
#define  SP_SET_SIG0             0x00400        // Bit 10: Set Signal 0
#define  SP_CLR_SIG1             0x00800        // Bit 11: Clear Signal 1
#define  SP_SET_SIG1             0x01000        // Bit 12: Set Signal 1
#define  SP_CLR_SIG2             0x02000        // Bit 13: Clear Signal 2
#define  SP_SET_SIG2             0x04000        // Bit 14: Set Signal 2
#define  SP_CLR_SIG3             0x08000        // Bit 15: Clear Signal 3
#define  SP_SET_SIG3             0x10000        // Bit 16: Set Signal 3
#define  SP_CLR_SIG4             0x20000        // Bit 17: Clear Signal 4
#define  SP_SET_SIG4             0x40000        // Bit 18: Set Signal 4
#define  SP_CLR_SIG5             0x80000        // Bit 19: Clear Signal 5
#define  SP_SET_SIG5             0x100000       // Bit 20: Set Signal 5
#define  SP_CLR_SIG6             0x200000       // Bit 21: Clear Signal 6
#define  SP_SET_SIG6             0x400000       // Bit 22: Set Signal 6
#define  SP_CLR_SIG7             0x800000       // Bit 23: Clear Signal 7
#define  SP_SET_SIG7             0x1000000      // Bit 24: Set Signal 7

// Patterns to Interpret Status Register (SP_STATUS_REG - Read)
 
#define  SP_STATUS_HALT          0x001          // Bit  0: Halt
#define  SP_STATUS_BROKE         0x002          // Bit  1: Broke
#define  SP_STATUS_DMA_BUSY      0x004          // Bit  2: DMA Busy
#define  SP_STATUS_DMA_FULL      0x008          // Bit  3: DMA Full
#define  SP_STATUS_IO_FULL       0x010          // Bit  4: IO Full
#define  SP_STATUS_SSTEP         0x020          // Bit  5: Single Step
#define  SP_STATUS_INTR_BREAK    0x040          // Bit  6: Interrupt on Break
#define  SP_STATUS_SIG0          0x080          // Bit  7: Signal 0 Set
#define  SP_STATUS_SIG1          0x100          // Bit  8: Signal 1 Set
#define  SP_STATUS_SIG2          0x200          // Bit  9: Signal 2 Set
#define  SP_STATUS_SIG3          0x400          // Bit 10: Signal 3 Set
#define  SP_STATUS_SIG4          0x800          // Bit 11: Signal 4 Set
#define  SP_STATUS_SIG5          0x1000         // Bit 12: Signal 5 Set
#define  SP_STATUS_SIG6          0x2000         // Bit 13: Signal 6 Set
#define  SP_STATUS_SIG7          0x4000         // Bit 14: Signal 7 Set

// Use of SIG Bits
 
#define  SP_CLR_YIELD            SP_CLR_SIG0
#define  SP_SET_YIELD            SP_SET_SIG0
#define  SP_STATUS_YIELD         SP_STATUS_SIG0
#define  SP_CLR_YIELDED          SP_CLR_SIG1
#define  SP_SET_YIELDED          SP_SET_SIG1
#define  SP_STATUS_YIELDED       SP_STATUS_SIG1
#define  SP_CLR_TASKDONE         SP_CLR_SIG2
#define  SP_SET_TASKDONE         SP_SET_SIG2
#define  SP_STATUS_TASKDONE      SP_STATUS_SIG2

// SP IMEM BIST REG
// READ/WRITE
// Bits [6:0]        BIST Status Bits (See Below for Detail)

#define  SP_IBIST_REG            0x04080004

// Patterns to Interpret Status Register (SP_BIST_REG - Write)
 
#define  SP_IBIST_CHECK          0x01     // Bit 0: BIST Check
#define  SP_IBIST_GO             0x02     // Bit 1: BIST Go
#define  SP_IBIST_CLEAR          0x04     // Bit 2: BIST Clear

// Patterns to Interpret Status Register (SP_BIST_REG - Read)
// First 2 Bits are Same as in Write Mode
// Bits [0]          BIST Check; Bit 1: BIST Go 
 
#define  SP_IBIST_DONE           0x04     // Bit 2: BIST Done
#define  SP_IBIST_FAILED         0x78     // Bit [6:3]: BIST Fail


////////////////////////////////////////////////////////////////////////////////
// DP Command Registers 
 
// DP CMD DMA Start
// READ/WRITE
// Bits [23:0]       DMEM/RDRAM Start Address

#define  DPC_START_REG           0x04100000

// DP CMD DMA End
// READ/WRITE
// Bits [23:0]       DMEM/RDRAM End Address

#define  DPC_END_REG             0x04100004

// DP CMD DMA End
// READ
// Bits [23:0]       DMEM/RDRAM Current Address

#define  DPC_CURRENT_REG         0x04100008

// DP CMD Status
// READ/WRITE
// Bits [9:0]        Valid Bits (See below for Definitions)

#define  DPC_STATUS_REG          0x0410000C

// DP Clock Counter
// READ
// Bits [23:0]       Clock Counter

#define  DPC_CLOCK_REG           0x04100010

// DP Buffer Busy Counter
// READ
// Bits [23:0]       Clock Counter

#define  DPC_BUFBUSY_REG         0x04100014

// DP Pipe Busy Counter
// READ
// Bits [23:0]       Clock Counter

#define  DPC_PIPEBUSY_REG        0x04100018

// DP TMEM Load Counter
// READ 
// Bits [23:0]       Clock Counter

#define  DPC_TMEM_REG            0x0410001C

// Values to Clear/Set Bit in Status Register (DPC_STATUS_REG - Write)

#define  DPC_CLR_XBUS_DMEM_DMA   0x0001      // Bit 0: Clear xbus_dmem_dma 
#define  DPC_SET_XBUS_DMEM_DMA   0x0002      // Bit 1: Set xbus_dmem_dma
#define  DPC_CLR_FREEZE          0x0004      // Bit 2: Clear Freeze
#define  DPC_SET_FREEZE          0x0008      // Bit 3: Set Freeze
#define  DPC_CLR_FLUSH           0x0010      // Bit 4: Clear Flush
#define  DPC_SET_FLUSH           0x0020      // Bit 5: Set Flush
#define  DPC_CLR_TMEM_CTR        0x0040      // Bit 6: Clear Tmem Ctr
#define  DPC_CLR_PIPE_CTR        0x0080      // Bit 7: Clear Pipe Ctr
#define  DPC_CLR_CMD_CTR         0x0100      // Bit 8: Clear Cmd Ctr
#define  DPC_CLR_CLOCK_CTR       0x0200      // Bit 9: Clear Clock Ctr

// Patterns to Interpret Status Register (DPC_STATUS_REG - Read)
 
#define  DPC_STATUS_XBUS_DMEM_DMA   0x001    // Bit  0: xbus_dmem_dma
#define  DPC_STATUS_FREEZE          0x002    // Bit  1: Freeze 
#define  DPC_STATUS_FLUSH           0x004    // Bit  2: Flush
#define  DPC_STATUS_START_GCLK      0x008    // Bit  3: Start Gclk
#define  DPC_STATUS_TMEM_BUSY       0x010    // Bit  4: Tmem Busy
#define  DPC_STATUS_PIPE_BUSY       0x020    // Bit  5: Pipe Busy
#define  DPC_STATUS_CMD_BUSY        0x040    // Bit  6: Cmd Busy
#define  DPC_STATUS_CBUF_READY      0x080    // Bit  7: Dbuf Ready
#define  DPC_STATUS_DMA_BUSY        0x100    // Bit  8: Dma Busy
#define  DPC_STATUS_END_VALID       0x200    // Bit  9: End Valid
#define  DPC_STATUS_START_VALID     0x400    // Bit 10: Start Valid


////////////////////////////////////////////////////////////////////////////////
// DP Span Registers 
 
// DP Tmem Bist 
// READ/WRITE
// Bits [10:0]    BIST Status Bits (See below for detail)

#define  DPS_TBIST_REG           0x04200000

// DP Span Test Mode
// READ/WRITE
// Bits [0]       Span Buffer Test Access Enable

#define  DPS_TEST_MODE_REG       0x04200004

// DP Span Buffer Test Address
// Bits [6:0]     See Below for Detail

#define  DPS_BUFTEST_ADDR_REG    0x04200008

// DP Span Buffer Test Data
// READ/WRITE
// Bits [31:0]    Span Buffer Data

#define  DPS_BUFTEST_DATA_REG    0x0420000C

// Patterns to Interpret Status Register (DPS_TMEM_BIST_REG - Write)
 
#define  DPS_TBIST_CHECK         0x01     // Bit 0: BIST Check
#define  DPS_TBIST_GO            0x02     // Bit 1: BIST Go
#define  DPS_TBIST_CLEAR         0x04     // Bit 2: BIST Clear

// Patterns to Interpret Status Register (DPS_TMEM_BIST_REG - Read)
// First 2 Bits are Same as in Write Mode
// Bits [0]       BIST Check
// Bits [1]       BIST Go 
 
#define  DPS_TBIST_DONE          0x004    // Bit 2: BIST Done
#define  DPS_TBIST_FAILED        0x7F8    // Bit [10:3]: BIST Fail


////////////////////////////////////////////////////////////////////////////////
// MIPS Interface (MI) Registers 

// MI Init Mode
// WRITE
// Bits [6:0]     Init Length
// Bits [7]       Clear Init Mode
// Bits [8]       Set Init Mode
// Bits [9/10]    Clear/Set Ebus Test Mode
// Bits [11]      Clear DP Interrupt
// READ
// Bits [6:0]     Init Length
// Bits [7]       Init Mode
// Bits [8]       Ebus Test Mode
 
#define  MI_INIT_MODE_REG        0x04300000

// Values to Clear/Set Bit in Mode Register (MI_MODE_REG - Write)
 
#define  MI_CLR_INIT             0x0080      // Bit  7: Clear Init Mode
#define  MI_SET_INIT             0x0100      // Bit  8: Set Init Mode
#define  MI_CLR_EBUS             0x0200      // Bit  9: Clear Ebus Test
#define  MI_SET_EBUS             0x0400      // Bit 10: Set Ebus Test Mode
#define  MI_CLR_DP_INTR          0x0800      // Bit 11: Clear DP Interrupt
#define  MI_CLR_RDRAM            0x1000      // Bit 12: Clear RDRAM Reg
#define  MI_SET_RDRAM            0x2000      // Bit 13: Set RDRAM Reg Mode

// Patterns to Interpret Mode Register (MI_MODE_REG - Read)
 
#define  MI_MODE_INIT            0x0080      // Bit  7: Init Mode
#define  MI_MODE_EBUS            0x0100      // Bit  8: Ebus Test Mode
#define  MI_MODE_RDRAM           0x0200      // Bit  9: RDRAM Reg Mode

// MI Version
// READ
// Bits [7:0]     IO
// Bits [15:8]    RAC
// Bits [23:16]   RDP
// Bits [31:24]   RSP

#define  MI_VERSION_REG          0x04300004

// MI Interrupt
// READ
// Bits [5:0]     Valid Bits (See Below for Bit Patterns)

#define  MI_INTR_REG             0x04300008

// MI Interrupt Mask
// WRITE
// Bits [11:0]    Valid Bits (See Below for Bit Patterns)
// READ
// Bits [5:0]     Valid Bits (See Below for Bit Patterns)
 
#define  MI_INTR_MASK_REG        0x0430000C

// The following are Values to Check for Interrupt Setting (MI_INTR_REG)
 
#define  MI_INTR_NO              0x00     // Clear
#define  MI_INTR_SP              0x01     // Bit 0: SP Interrupt
#define  MI_INTR_SI              0x02     // Bit 1: SI Interrupt
#define  MI_INTR_AI              0x04     // Bit 2: AI Interrupt
#define  MI_INTR_VI              0x08     // Bit 3: VI Interrupt
#define  MI_INTR_PI              0x10     // Bit 4: PI Interrupt
#define  MI_INTR_DP              0x20     // Bit 5: DP Interrupt

// The following are Values to Clear/Set Various Interrupt Bit Mask
// They can be ORed together to Manipulate Multiple Bits 
// (MI_INTR_MASK_REG - Write)
 
#define  MI_INTR_MASK_CLR_SP     0x0001   // Bit  0: Clear SP Mask
#define  MI_INTR_MASK_SET_SP     0x0002   // Bit  1: Set SP Mask
#define  MI_INTR_MASK_CLR_SI     0x0004   // Bit  2: Clear SI Mask
#define  MI_INTR_MASK_SET_SI     0x0008   // Bit  3: Set SI Mask
#define  MI_INTR_MASK_CLR_AI     0x0010   // Bit  4: Clear AI Mask
#define  MI_INTR_MASK_SET_AI     0x0020   // Bit  5: Set AI Mask
#define  MI_INTR_MASK_CLR_VI     0x0040   // Bit  6: Clear VI Mask
#define  MI_INTR_MASK_SET_VI     0x0080   // Bit  7: Set VI Mask 
#define  MI_INTR_MASK_CLR_PI     0x0100   // Bit  8: Clear PI Mask
#define  MI_INTR_MASK_SET_PI     0x0200   // Bit  9: Set PI Mask
#define  MI_INTR_MASK_CLR_DP     0x0400   // Bit 10: Clear DP Mask
#define  MI_INTR_MASK_SET_DP     0x0800   // Bit 11: Set DP Mask

// The following are Values to Check for Interrupt Mask Setting 
// (MI_INTR_MASK_REG - Read)
 
#define  MI_INTR_MASK_SP         0x01     // Bit 0: SP Interrupt Mask
#define  MI_INTR_MASK_SI         0x02     // Bit 1: SI Interrupt Mask
#define  MI_INTR_MASK_AI         0x04     // Bit 2: AI Interrupt Mask
#define  MI_INTR_MASK_VI         0x08     // Bit 3: VI Interrupt Mask
#define  MI_INTR_MASK_PI         0x10     // Bit 4: PI Interrupt Mask
#define  MI_INTR_MASK_DP         0x20     // Bit 5: DP Interrupt Mask


////////////////////////////////////////////////////////////////////////////////
// Video Interface (VI) Registers 
 
#define  VI_INTR_TIME            0x98968

// VI Status/Control
// READ/WRITE
// Bits [15-0]    Valid Bits
// Bits [1:0]     Type [1:0] (Pixel Size) 
//                0: Blank (No Data, No Sync)
//                1: Reserved
//                2: 5/5/5/3 (16 Bit)
//                3: 8/8/8/8 (32 Bit)
// Bits [2]       Gamma_Dither_Enable (Normally On, unless "Special Effect")
// Bits [3]       Gamma_Enable (Normally On, unless MPEG/JPEG)
// Bits [4]       Divot_Enable (Normally On if Antialiased, unless Decal Lines)
// Bits [5]       Reserved - Always Off
// Bits [6]       Serrate (Always on if Interlaced, Off if Not)
// Bits [7]       Reserved - Diagnostics Only
// Bits [9:8]     Anti-Alias (AA) Mode [1:0] 
//                0: AA & Resamp (Always Fetch Extra Lines)
//                1: AA & Resamp (Fetch Extra Lines if Needed)
//                2: Resamp Only (Treat as all Fully Covered)
//                3: Neither (Replicate Pixels, No Interpolate)
// Bits [11]      Reserved - Diagnostics Only
// Bits [15:12]   Reserved

#define  VI_STATUS_REG           0x04400000

// VI Origin
// READ/WRITE
// Bits [23:0]    Frame Buffer Origin in Bytes

#define  VI_ORIGIN_REG           0x04400004

// VI Width 
// READ/WRITE
// Bits [11:0]    Frame Buffer Line Width in Pixels

#define  VI_WIDTH_REG            0x04400008

// VI Vertical Interrupt
// READ/WRITE
// Bits [9:0]     Interrupt when Current Half-line = V_INTR

#define  VI_INTR_REG             0x0440000C

// VI Current Vertical Line
// READ/WRITE
// Bits [9:0]     Current Half Line, Sampled Once per Line (the LSB of V_CURRENT
//                is Constant within a Field, and in Interlaced Modes gives the
//                Field Number - which is Constant for Non-interlaced Modes)
//                Any Write to this Register will Clear Interrupt Line

#define  VI_CURRENT_REG          0x04400010

// VI Video Timing
// READ/WRITE
// Bits [7:0]     Horizontal Sync Width in Pixels
// Bits [15:8]    Color Burst Width in Pixels
// Bits [19:16]   Vertical Sync Width in Half Lines,   
// Bits [29:20]   Start of Color Burst in Pixels from H-Sync
 
#define  VI_BURST_REG            0x04400014

// VI Vertical Sync
// READ/WRITE
// Bits [9:0]     Number of Half-Lines per Field

#define  VI_V_SYNC_REG           0x04400018

// VI Horizontal Sync
// READ/WRITE
// Bits [11:0]    Total Duration of a Line in 1/4 Pixel 
// Bits [20:16]   A 5-Bit Leap Pattern used for PAL Only 
//                (H_Sync_Period)
 
#define  VI_H_SYNC_REG           0x0440001C

// VI Horizontal Sync Leap 
// READ/WRITE
// Bits [11:0]    Identical to H_Sync_Period
// Bits [27:16]   Identical to H_Sync_Period
 
#define  VI_LEAP_REG             0x04400020

// VI Horizontal Video
// READ/WRITE
// Bits [9:0]     End of Active Video in Screen Pixels
// Bits [25:16]   Start of Active Video in Screen Pixels

#define  VI_H_START_REG          0x04400024

// VI Vertical Video
// READ/WRITE
// Bits [9:0]     End of Active Video in Screen Half-Lines
// Bits [25:16]   Start of Active Video in Screen Half-Lines
 
#define  VI_V_START_REG          0x04400028

// VI Vertical Burst
// READ/WRITE
// Bits [9:0]     End of Color Burst Enable in Half-Lines
// Bits [25:16]   Start of Color Burst Enable in Half-Lines

#define  VI_V_BURST_REG          0x0440002C

// VI X-Scale
// READ/WRITE
// Bits [11:0]    1/Horizontal Scale up Factor (2.10 Format)
// Bits [27:16]   Horizontal Subpixel Offset (2.10 Format)
 
#define  VI_X_SCALE_REG          0x04400030

// VI Y-Scale
// READ/WRITE
// Bits [11:0]    1/Vertical Scale up Factor (2.10 Format)
// Bits [27:16]   Vertical Subpixel Offset (2.10 Format)
 
#define  VI_Y_SCALE_REG          0x04400034

// Patterns to Interpret VI_CONTROL_REG
 
#define  VI_CTRL_TYPE_16            0x00002  // Bit [1:0] Pixel Size: 16 Bit
#define  VI_CTRL_TYPE_32            0x00003  // Bit [1:0] Pixel Size: 32 Bit
#define  VI_CTRL_GAMMA_DITHER_ON    0x00004  // Bit 2: Default = On
#define  VI_CTRL_GAMMA_ON           0x00008  // Bit 3: Default = On
#define  VI_CTRL_DIVOT_ON           0x00010  // Bit 4: Default = On
#define  VI_CTRL_SERRATE_ON         0x00040  // Bit 6: On if Interlaced
#define  VI_CTRL_ANTIALIAS_MASK     0x00300  // Bit [9:8] Anti-Alias Mode
#define  VI_CTRL_DITHER_FILTER_ON   0x10000  // Bit 16: Dither-Filter Mode

// Possible Video Clocks (NTSC or PAL)
 
#define  VI_NTSC_CLOCK           48681812    // Hz = 48.681812 MHz
#define  VI_PAL_CLOCK            49656530    // Hz = 49.656530 MHz
#define  VI_MPAL_CLOCK           48628316    // Hz = 48.628316 MHz


////////////////////////////////////////////////////////////////////////////////
// Audio Interface (AI) Registers 
//
// The Address and Length Registers are Double Buffered; that is, they
// can be Written Twice before becoming Full. The Address must be Written 
// before the Length.
 
// AI DRAM Address
// WRITE
// Bits [23:0]    Starting RDRAM Address (8B-aligned)

#define  AI_DRAM_ADDR_REG        0x04500000     // R0: DRAM Address

// AI Length
// READ/WRITE
// Bits [14:0]    Transfer Length (v1.0) - Bottom 3 Bits are Ignored
// Bits [17:0]    Transfer Length (v2.0) - Bottom 3 Bits are Ignored

#define  AI_LEN_REG              0x04500004     // R1: Length

// AI Control
// WRITE
// Bits [0]       DMA Enable - If LSB == 1, DMA is enabled

#define  AI_CONTROL_REG          0x04500008     // R2: DMA Control 

// AI Status
// READ
// Bits [31]/[0]  AI_Full (Addr & Len Buffer Full)
// Bits [30]      AI_Busy
// Note that a 1->0 Transition in AI_Full will set Interrupt
// WRITE
// Clear Audio Interrupt 
 
#define  AI_STATUS_REG           0x0450000C     // R3: Status

// AI DAC Sample Period Register
// WRITE
// Bits [13:0]    DAC Rate 
//                Vid_Clock/(Dperiod + 1) is the DAC Sample Rate
//                (Dperiod + 1) >= 66 * (Aclockhp + 1) must be True
 
#define  AI_DACRATE_REG          0x04500010     // R4: DAC rate 14-LSB

// AI Bit Rate
// WRITE
// Bits [3:0]     Bit Rate (Abus Clock Half Period Register - Aclockhp)
//                Vid_Clock/(2 * (Aclockhp + 1)) is the DAC Clock Rate
//                The Abus Clock Stops if Aclockhp is Zero
 
#define  AI_BITRATE_REG          0x04500014     // R5: Bit Rate 4-LSB

// Value for Control Register

#define  AI_CONTROL_DMA_ON       0x01           // LSB = 1: DMA Enable
#define  AI_CONTROL_DMA_OFF      0x00           // LSB = 1: DMA Enable

// Value for Status Register

#define  AI_STATUS_FIFO_FULL     0x80000000     // Bit 31: Full
#define  AI_STATUS_DMA_BUSY      0x40000000     // Bit 30: Busy

// DAC Rate = Video Clock / Audio Frequency
// DAC Rate >= (66 * Bit Rate) must be True
 
#define  AI_MAX_DAC_RATE         16384          // 14-Bit+1
#define  AI_MIN_DAC_RATE         132

// Bit Rate <= (DAC Rate / 66)

#define  AI_MAX_BIT_RATE         16             // 4-Bit+1
#define  AI_MIN_BIT_RATE         2

// Maximum and Minimum Values for Audio Frequency based on Video Clocks
// Max Frequency = (Video Clock / Min DAC Rate)
// Min Frequency = (Video Clock / Max DAC Rate)

#define  AI_NTSC_MAX_FREQ        368000          // 368 KHz
#define  AI_NTSC_MIN_FREQ        3000            //   3 KHz ~ 2971 Hz

#define  AI_PAL_MAX_FREQ         376000          // 376 KHz
#define  AI_PAL_MIN_FREQ         3050            //   3 KHz ~ 3031 Hz

#define  AI_MPAL_MAX_FREQ        368000          // 368 KHz
#define  AI_MPAL_MIN_FREQ        3000            //   3 KHz ~ 2968 Hz


////////////////////////////////////////////////////////////////////////////////
// Peripheral Interface (PI) Registers 
 
// PI DRAM Address
// READ/WRITE
// Bits [23:0]    Starting RDRAM Address

#define  PI_DRAM_ADDR_REG        0x04600000     // DRAM Address 

// PI Pbus (Cartridge) Address
// READ/WRITE
// Bits [31:0]    Starting AD16 Address

#define  PI_CART_ADDR_REG        0x04600004

// PI Read Length
// READ/WRITE
// Bits [23:0]    Read Data Length

#define  PI_RD_LEN_REG           0x04600008

// PI Write Length
// READ/WRITE
// Bits [23:0]    Write Data Length

#define  PI_WR_LEN_REG           0x0460000C

// PI Status
// READ
// Bits [0]       DMA Busy
// Bits [1]       IO Busy
// Bits [2]       Error
// WRITE
// Bits [0]       Reset Controller (and Abort Current Op)
// Bits [1]       Clear Interrupt
 
#define  PI_STATUS_REG           0x04600010

// PI DOM1 Latency
// READ/WRITE
// Bits [7:0]     Domain 1 Device Latency 

#define  PI_BSD_DOM1_LAT_REG     0x04600014

// PI DOM1 Pulse Width
// READ/WRITE
// Bits [7:0]     Domain 1 Device R/W Strobe Pulse Width

#define  PI_BSD_DOM1_PWD_REG     0x04600018

// PI DOM1 Page Size
// READ/WRITE 
// Bits [3:0]     Domain 1 Device Page Size

#define  PI_BSD_DOM1_PGS_REG     0x0460001C

// PI DOM1 Release
// READ/WRITE
// Bits [1:0]     Domain 1 Device R/W Release Duration 

#define  PI_BSD_DOM1_RLS_REG     0x04600020

// PI DOM2 latency
// READ/WRITE
// Bits [7:0]     Domain 2 Device Latency

#define  PI_BSD_DOM2_LAT_REG     0x04600024

// PI DOM2 Pulse Width
// READ/WRITE 
// Bits [7:0]     Domain 2 Device R/W Strobe Pulse Width

#define  PI_BSD_DOM2_PWD_REG     0x04600028

// PI DOM2 Page Size
// READ/WRITE
// Bits [3:0]     Domain 2 Device Page Size

#define  PI_BSD_DOM2_PGS_REG     0x0460002C

// PI DOM2 Release 
// READ/WRITE
// Bits [1:0]     Domain 2 Device R/W Release Duration

#define  PI_BSD_DOM2_RLS_REG     0x04600030

#define  PI_DOMAIN1_REG          PI_BSD_DOM1_LAT_REG
#define  PI_DOMAIN2_REG          PI_BSD_DOM2_LAT_REG

#define  PI_DOM_LAT_OFS          0x00
#define  PI_DOM_PWD_OFS          0x04
#define  PI_DOM_PGS_OFS          0x08
#define  PI_DOM_RLS_OFS          0x0C

// PI Status Register has 3 Bits Active when Read from (PI_STATUS_REG - Read)
// Bit 0: DMA Busy - Set when DMA is in Progress
//	Bit 1: IO Busy  - Set when IO is in Progress
// Bit 2: Error    - Set when CPU Issues IO Request while DMA is Busy
 
#define  PI_STATUS_ERROR         0x04
#define  PI_STATUS_IO_BUSY       0x02
#define  PI_STATUS_DMA_BUSY      0x01

// PI Status Register has 2 Bits Active when Written to:
//	Bit 0: When Set, Reset PIC
//	Bit 1: When Set, Clear Interrupt Flag
// The values of the two bits can be ORed together to both reset PIC and 
// clear interrupt at the same time.
//
// Note: 
//	The PIC does generate an interrupt at the end of each DMA. CPU 
//	needs to clear the interrupt flag explicitly (from an interrupt 
//	handler) by writing into the STATUS register with bit 1 set.
//
//	When a DMA completes, the interrupt flag is set.  CPU can issue
//	another request even while the interrupt flag is set (as long as
//	PIC is idle).  However, it is the CPU's responsibility for
//	maintaining accurate correspondence between DMA completions and
//	interrupts.
//
//	When PIC is reset, if PIC happens to be busy, an interrupt will
//	be generated as PIC returns to idle.  Otherwise, no interrupt will
//	be generated and PIC remains idle.

// Values to Clear Interrupt/Reset PIC (PI_STATUS_REG - Write)
 
#define  PI_STATUS_RESET         0x01

#define  PI_STATUS_CLR_INTR      0x02

#define  PI_DMA_BUFFER_SIZE      128

#define  PI_DOM1_ADDR1           0x06000000     /* To 0x07FFFFFF */
#define  PI_DOM1_ADDR2           0x10000000     /* To 0x1FBFFFFF */
#define  PI_DOM1_ADDR3           0x1FD00000     /* To 0x7FFFFFFF */
#define  PI_DOM2_ADDR1           0x05000000     /* To 0x05FFFFFF */
#define  PI_DOM2_ADDR2           0x08000000     /* To 0x0FFFFFFF */


////////////////////////////////////////////////////////////////////////////////
// RDRAM Interface (RI) Registers 
 
// RI Mode 
// READ/WRITE
// Bits [1:0]     Operating Mode
// Bits [2]       Stop T Active
// Bits [3]       Stop R Active 

#define  RI_MODE_REG             0x04700000

// RI Config
// READ/WRITE
// Bits [5:0]     Current Control Input
// Bits [6]       Current Control Enable

#define  RI_CONFIG_REG           0x04700004

// RI Current Load
// WRITE
// Bits []        Any Write Updates Current Control Register

#define  RI_CURRENT_LOAD_REG     0x04700008

// RI Select
// READ/WRITE
// Bits [2:0]     Receive Select
// Bits [2:0]     Transmit Select

#define  RI_SELECT_REG           0x0470000C

// RI Refresh
// READ/WRITE
// Bits [7:0]     Clean Refresh Delay
// Bits [15:8]    Dirty Refresh Delay,
// Bits [16]      Refresh Bank
// Bits [17]      Refresh Enable 
// Bits [18]      Refresh Optimize 
 
#define  RI_REFRESH_REG          0x04700010

// RI Latency
// READ/WRITE
// Bits [3:0]     DMA Latency/Overlap 

#define  RI_LATENCY_REG          0x04700014

// RI Error
// READ
// Bits [0]       NACK Error
// Bits [1]       ACK Error

#define  RI_RERROR_REG           0x04700018

// RI Error
// WRITE
// Bits []        Any Write Clears all Error Bits

#define  RI_WERROR_REG           0x0470001C


////////////////////////////////////////////////////////////////////////////////
// Serial Interface (SI) Registers 
 
// SI DRAM Address
// READ/WRITE
// Bits [23:0]    Starting RDRAM Address

#define  SI_DRAM_ADDR_REG        0x04800000

// SI Address Read 64B
// WRITE
// Bits []        Any Write Causes a 64B DMA Write

#define  SI_PIF_ADDR_RD64B_REG   0x04800004

// Address SI_BASE_REG + (0x08, 0x0C, 0x14) are Reserved

// SI Address Write 64B
// WRITE
// Bits []        Any Write causes a 64B DMA Read

#define  SI_PIF_ADDR_WR64B_REG   0x04800010

// SI Status
// WRITE
// Bits []        Any Write Clears Interrupt
// READ
// Bits [0]       DMA Busy
// Bits [1]       IO Read Busy
// Bits [2]       Reserved
// Bits [3]       DMA Error
// Bits [12]      Interrupt
 
#define  SI_STATUS_REG           0x04800018

// SI Status Register has the following Bits Active:
// 0:DMA busy		- Set when DMA is in Progress
//	1:IO busy		- Set when IO access is in Progress
//	3:DMA error		- Set when there are Overlapping DMA Requests
// 12:Interrupt	- Interrupt set
 
#define  SI_STATUS_DMA_BUSY      0x0001
#define  SI_STATUS_RD_BUSY       0x0002
#define  SI_STATUS_DMA_ERROR     0x0008
#define  SI_STATUS_INTERRUPT     0x1000


#endif  /* _RCP_H_ */

